// Title:   BilateralDenoiser.cpp
// Created: Thu Sep 25 15:11:54 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "BilateralDenoiser.h"


inline float BilateralDenoiser::gaussianSquared(float r, float sigma) {
	return (exp (-r / (2 * sigma * sigma)));
}


void BilateralDenoiser::filter(int curPosIndex, Vector3D *weightedPos, float *k, float spatial_sigma, float influence_sigma, bool mollify) {

	Vector3D pos, p, v;
	float w = 0.0f;

	// surfelArea could also be integrated to take surfel size into account
	// float surfelArea;

	Vector3D curPos = this->surfelPositions[curPosIndex];
	nbHood->setSourcePoint(curPos);
	nbHood->setNofNeighbours(3);
	float minSquaredDistance = Vector3D(this->nbHood->getNeighbourPosition(1) - this->nbHood->getNeighbourPosition(0)).getSquaredLength();
	if (minSquaredDistance > (cutoff * cutoff)) {
		return;
	}

	// alternatively a query grid could be used. but because kd-tree queries are faster
	// we limit the number of neighbours to a certain amount.
	int nofNeighbours = 100;
	this->nbHood->setNofNeighbours(nofNeighbours);

	for (int j = 1; j < nofNeighbours; j++) {

		pos = this->nbHood->getNeighbourPosition(j);

		// check if outside spatial influence radius
		if (Vector3D::distance(curPos, pos) > spatial_sigma*this->meanLength) {
			break;
		}

		Vector3D norm;
		if (mollify) {
			p = pos;
		}
		else {
			norm = this->newSurfelNormals[this->nbHood->getNeighbourPositionIndex(j)];
			p = Vector3D::projectPointOntoPlane(curPos, pos, norm);
		}

		if (influence_sigma > 0.0) {
			float squaredDist = Vector3D(p - curPos).getSquaredLength();
			float spatialDist = Vector3D(curPos - pos).getSquaredLength();
			// w += surfelArea * gaussianSquared(spatialDist, spatial_sigma) * gaussianSquared(squaredDist, influence_sigma);
			w += gaussianSquared(spatialDist, spatial_sigma) * gaussianSquared(squaredDist, influence_sigma);
		}
		else {
			// w += surfelArea * spatial_sigma;
			w += spatial_sigma;
		}
			*weightedPos += (p * w);
			*k += w;
	}

	return;
}


void BilateralDenoiser::mollifyVertex(int curPosIndex) {

	float k = 0.0f;
	Vector3D newPos = Vector3D(0.0f, 0.0f, 0.0f);
	filter(curPosIndex, &newPos, &k, this->sigma_m, this->sigma_g, true);

	if (k > 0.0) {
		Vector3D normal = Vector3D((newPos / k) - this->surfelPositions[curPosIndex]);
		normal.normalize();
		if (Vector3D::dotProduct(normal, this->surfelNormals[curPosIndex]) < 0.0f) {
			normal *= -1;
		}
		this->newSurfelNormals[curPosIndex] = normal;
	}
	else {
		this->newSurfelNormals[curPosIndex] = Vector3D(this->surfelNormals[curPosIndex]);
	}

	return;
}


void BilateralDenoiser::filterVertex(int curPosIndex) {

	float k = 0.0f;
	Vector3D newPos = Vector3D(0.0f, 0.0f, 0.0f);
	filter(curPosIndex, &newPos, &k, this->sigma_m, this->sigma_g, false);
	if (k > 0.0) {
		this->newSurfelPositions[curPosIndex] = Vector3D(newPos / k);
	}
	else {
		this->newSurfelPositions[curPosIndex] = Vector3D(this->surfelPositions[curPosIndex]);
	}

	return;
}


void BilateralDenoiser::smooth(float sigma_f, float sigma_g) {

	Scene			 *scene;
	Object			 *object;
	SurfelCollection *surfels;
	SurfelInterface	 *surfel;
	int				 i;

	// copy parameters
	this->sigma_f = sigma_f;
	this->sigma_g = sigma_g;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	bool isNumber = false;

	if (object) {
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();

		if (nofSurfels == 0) {
			return;
		}

		this->surfelPositions = new Vector3D [nofSurfels];
		this->newSurfelPositions = new Vector3D [nofSurfels];
		this->surfelNormals = new Vector3D [nofSurfels];
		this->newSurfelNormals = new Vector3D [nofSurfels];

		// build up structure
		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			this->surfelPositions[i] = surfel->getPosition();
			this->surfelNormals[i] = surfel->getNormal();
		}

		this->nbHood = new NeighbourHood(this->surfelPositions, nofSurfels);

		// calculate mean length between surfels
		for (i = 0; i < nofSurfels; i++) {
			this->nbHood->setNofNeighbours(2);
			this->nbHood->setSourcePoint(this->surfelPositions[i]);
			meanLength += Vector3D(this->nbHood->getNeighbourPosition(1) - this->nbHood->getNeighbourPosition(0)).getLength();
		}
		
		this->meanLength /= (nofSurfels);
		
		// spatial and influence sigma
		this->sigma_f *= meanLength;
		this->sigma_g *= meanLength;
		// sigma for mollification
		this->sigma_m = this->sigma_f / 2.0f;
		
		this->cutoff = 2.0f * sigma_f;

		for (i = 0; i < nofSurfels; i++) {
			// eventually mollification with a different method could improve results (e.g. Principal Component Analysis)
			mollifyVertex(i);
		}

		for (i = 0; i < nofSurfels; i++) {
			filterVertex(i);
		}

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			scene->setSurfelPosition(surfel, this->newSurfelPositions[i], false);
//			scene->setSurfelNormal(surfel, this->newSurfelNormals[i], false);
		}

		scene->emitObjectModifiedSignal();
		scene->emitRenderFrameSignal();
	
	}
}
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
