// Title:   Diffuser_DialogImpl.cpp
// Created: Thu Sep 25 15:11:45 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "Diffuser_DialogImpl.h"

// ******
// global
// ******

// the single Diffuser_DialogImpl instance, only
// accesible by getInstance() and
// destroyInstance()
static Diffuser_DialogImpl *instance = 0;

// *******************
// private c'tor/d'tor
// *******************

Diffuser_DialogImpl::Diffuser_DialogImpl (QWidget *parent, const char *name, bool modal, WFlags flags)
    : Diffuser_DialogBase (parent, name, modal, flags) {

	this->initDialog();

}


Diffuser_DialogImpl::~Diffuser_DialogImpl() {
    // no need to delete child widgets, Qt does it all for us
}

// **************
// public methods
// **************

Diffuser_DialogImpl *Diffuser_DialogImpl::getInstance() {

	if (instance == 0) {
		instance = new Diffuser_DialogImpl();
	}
	return instance;

}

void Diffuser_DialogImpl::destroyInstance() {

	if (instance != 0) {
		instance->close();
		delete instance;
	}
	instance = 0;

}

// ***************
// private methods
// ***************

void Diffuser_DialogImpl::initDialog() {

	// get informed whenever the number of objects in the scene changes
	this->connect (Scene::getInstance(), SIGNAL (objectAdded()),
		           this, SLOT (handleObjectAdded()));

	// get informed whenever the number of objects in the scene changes
	this->connect (Scene::getInstance(), SIGNAL (objectRemoved()),
		           this, SLOT (handleObjectRemoved()));
	

	// to disable the extra menu
	// this->DiffuserTabWidget->setTabEnabled(this->tab_5, false);

	this->grid = NULL;
	this->narrowBand = NULL;
	this->actualPixmap = NULL;
	// update maxDepth
	handle_ResolutionChanged(this->ResolutionComboBox->currentText());

	if (this->XYPlaneRadioButton->isChecked()) {
		this->plane = XZ_PLANE;
	}
	else if (this->YZPlaneRadioButton->isChecked()) {
		this->plane = YZ_PLANE;
	}
	else {
		this->plane = XY_PLANE;
	}

	this->reader = NULL;
	this->distanceFieldCalcMethod = TANGENTPLANES;
	this->particleSim = NULL;
	polynomial = 3;
}


////////////////////////////////////////////////////////////
// Preprocessing
////////////////////////////////////////////////////////////


void Diffuser_DialogImpl::handle_BilateralSmoothing() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	
	bool isNumber; 
	float spatialSigma, influenceSigma;

	// validate input
	spatialSigma = this->InfluenceSigmaLineEdit->text().toFloat(&isNumber);
	if (!isNumber) {
		this->InfluenceSigmaLineEdit->setText(QString("1.0"));
		spatialSigma = 1.0;
	}
	influenceSigma = this->SpatialSigmaLineEdit->text().toFloat(&isNumber);
	if (!isNumber) {
		this->SpatialSigmaLineEdit->setText(QString("1.0"));
		influenceSigma = 1.0;
	}

	BilateralDenoiser *denoiser = new BilateralDenoiser();
	denoiser->smooth(spatialSigma, influenceSigma);
	delete denoiser;

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_SmoothNormals() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface	 *surfel;
	int				 i;
	int				 numberOfNeighbors;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();
		
		// get number of neighbors
		numberOfNeighbors = this->SmoothNormalsSpinBox->value();

		if (nofSurfels <= numberOfNeighbors) {
			QApplication::restoreOverrideCursor();
			return;
		}
	
		AdjacencyGraph *graph = new AdjacencyGraph(surfels, numberOfNeighbors);
		
		for (surfel = surfels->getFirstSurfel(), i = 0; surfel; surfel = surfels->getNextSurfel(), i++) {
			surfel->setNormal(graph->computeNormal(i, numberOfNeighbors));
		}
		
		delete graph;
	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_RemoveCorruptSurfels() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;

	// Get the active object
	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		surfels = object->getSurfelCollection();

		for (surfel = surfels->getFirstSurfel(); surfel; surfel = surfels->getNextSurfel()) {
			if (surfels->getPropertyDescriptor() & SurfelInterface::TANGENT_AXES) {
				if (surfel->getAxisOne().getSquaredLength() == 0 || surfel->getAxisTwo().getSquaredLength() == 0) {
					surfels->removeCurrentSurfel();
					surfel = surfels->getCurrentSurfel();
				}
				else {
					surfel = surfels->getNextSurfel();
				}
			}
			else {
				if (surfel->getRadius() == 0.0) {
					surfels->removeCurrentSurfel();
					surfel = surfels->getCurrentSurfel();
				}
				else {
					surfel = surfels->getNextSurfel();
				}
			}
		}
		
	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_CenterOfMass() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	NeighbourHood	 *nbHood;
	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;
	Vector3D		 sourcePos;
	Vector3D		 centerOfMass;
	Vector3D		 *surfelPositions;
	float			 param;
	int				 i, j;
	int				 numberOfNeighbors;

	// Get the active object
	scene = Scene::getInstance();
	object = scene->getActiveObject();

	bool isNumber; 

	// validate input
	param = this->CoMToleranceLineEdit->text().toFloat(&isNumber);
	if (!isNumber) {
		this->CoMToleranceLineEdit->setText(QString("1.0"));
		param = 1.0;
	}
	numberOfNeighbors = this->CoMNumberOfNeighborsSpinBox->value();


	if (object) {
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();

		if (nofSurfels < numberOfNeighbors) {
			QApplication::restoreOverrideCursor ();
			return;
		}

		surfelPositions = new Vector3D [nofSurfels];

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			surfelPositions[i] = surfel->getPosition();
		}

		nbHood = new NeighbourHood (surfelPositions, nofSurfels);
		nbHood->setNofNeighbours(numberOfNeighbors);

		float meanLength = 0.0f;
		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			sourcePos = surfelPositions[i];
			nbHood->setSourcePoint(sourcePos);
			meanLength += Vector3D::distance(sourcePos, nbHood->getNeighbourPosition(1));
		}
		
		meanLength /= nofSurfels;

		SurfelInterface::PropertyDescriptor pd = surfels->getPropertyDescriptor() | SurfelInterface::FLAGS;
		surfels->setPropertyDescriptor(pd);
		
		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			sourcePos = surfelPositions[i];
			nbHood->setSourcePoint(sourcePos);
			
			centerOfMass.makeZero();
			for (j = 0; j < numberOfNeighbors; j++) {
				centerOfMass += nbHood->getNeighbourPosition(j);
			}
			centerOfMass /= numberOfNeighbors;

			// set selection flags
			if (Vector3D::distance(sourcePos, centerOfMass) >= param*meanLength) {
				scene->setSurfelFlagOn (surfel, (SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE), true, false);
			}
			else {
				scene->setSurfelFlagOn (surfel, (SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE), false, false);
			}
		}
	}

	delete nbHood;
	delete []surfelPositions;

	scene->emitRenderFrameSignal();
	scene->emitObjectModifiedSignal();

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_CenterOfMassRemoveSelected() {
	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;

	// Get the active object
	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		scene->removeSurfels((SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE), true);
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_FlipAllNormals() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;
	int i;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			scene->setSurfelNormal(surfel, -surfel->getNormal(), false);
		}
	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_ToleranceValueLineEditChanged(const QString& string) {
	QApplication::setOverrideCursor (Qt::waitCursor);

	bool isNumber; 

	float value = this->ToleranceValueLineEdit->text().toFloat(&isNumber);
	if (!isNumber || value < 0.0 || value > 2.0) {
		this->ToleranceValueLineEdit->setText(QString("1.0"));
	}
	else {
		this->ToleranceValueSlider->setValue(value * 1000);
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_ToleranceValueSliderChanged(int value) {
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	QString val;
	val.setNum(0.001f * value);
	this->ToleranceValueLineEdit->setText(val);
	
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_ClassifyBadNormals() {

	QApplication::setOverrideCursor (Qt::waitCursor);	

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;
	int				 i, j;
	Vector3D		 maximum, minimum;
	Vector3D		 center;
	int				 startingSurfels[] = {-1, -1, -1, -1, -1, -1};
	int				 nofNeighbours;
	int				 index;
	Vector3D		 newNormal, goodNormal, normal[6];
	int				 nofPlanes;
	Vector3D		 *planeNormals;
	int				 *planeIndices;
	float			 *planeMaxDist;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {

		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();

		SurfelInterface::PropertyDescriptor pd = surfels->getPropertyDescriptor() | SurfelInterface::FLAGS;
		surfels->setPropertyDescriptor(pd);

		if (nofSurfels < 6) {
			return;
		}

		if (AdditionalStartingSurfelsCheckBox->isChecked()) {
			nofPlanes = this->NumberOfPlanesSpinBox->value();;
			planeNormals = new Vector3D [nofPlanes];
			planeIndices = new int [nofPlanes];
			planeMaxDist = new float [nofPlanes];

			for (i = 0; i < nofPlanes; i++) {
				planeNormals[i] = Vector3D(rand(), rand(), rand());
				planeNormals[i].normalize();
				planeIndices[i] = -1;
				planeMaxDist[i] = 0.0f;
			}
		}

		nofNeighbours = this->FlippedNormalsNumberOfNeighborsSpinBox->value();

		Vector3D *surfelPositions = new Vector3D [nofSurfels];
		Vector3D *surfelNormals = new Vector3D [nofSurfels];
		std::set<int> okSurfels;
		std::set<int> surfelsToExpand;

		minimum = maximum = surfels->getFirstSurfel()->getPosition();

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			surfelPositions[i] = surfel->getPosition();
			surfelNormals[i] = surfel->getNormal();

			if (maximum[0] < surfelPositions[i][0]) {
				startingSurfels[0] = i;
				maximum[0] = surfelPositions[i][0];
			}
			if (maximum[1] < surfelPositions[i][1]) {
				startingSurfels[1] = i;
				maximum[1] = surfelPositions[i][1];
			}
			if (maximum[2] < surfelPositions[i][2]) {
				startingSurfels[2] = i;
				maximum[2] = surfelPositions[i][2];
			}
			if (minimum[0] > surfelPositions[i][0]) {
				startingSurfels[3] = i;
				minimum[0] = surfelPositions[i][0];
			}
			if (minimum[1] > surfelPositions[i][1]) {
				startingSurfels[4] = i;
				minimum[1] = surfelPositions[i][1];
			}
			if (minimum[2] > surfelPositions[i][2]) {
				startingSurfels[5] = i;
				minimum[2] = surfelPositions[i][2];
			}
		}

		if (AdditionalStartingSurfelsCheckBox->isChecked()) {

			center = Vector3D((minimum[0] + maximum[0]) / 2, (minimum[1] + maximum[1]) / 2, (minimum[2] + maximum[2]) / 2);

			for (i = 0; i < nofSurfels; i++) {
				for (j = 0; j < nofPlanes; j++) {
					float distanceToPlane = fabs(Vector3D::dotProduct(planeNormals[j], surfelPositions[i] - center));
					if (distanceToPlane >= planeMaxDist[j]) {
						planeMaxDist[j] = distanceToPlane;
						planeIndices[j] = i;
					}
				}			
			}
			if (AdditionalStartingSurfelsCheckBox->isChecked()) {
				for (i = 0; i < nofSurfels; i++) {
					for (j = 0; j < nofPlanes; j++) {
						surfelNormals[planeIndices[j]] = planeNormals[j]; 
					}			
				}
			}
		}

		MinPriorityQueue<Edge*, float> heap;
		heap.setSize(nofSurfels*nofNeighbours);
		NeighbourHood search = NeighbourHood (surfelPositions, nofSurfels);

		normal[0] = Vector3D( 1, 0, 0);
		normal[1] = Vector3D( 0, 1, 0);
		normal[2] = Vector3D( 0, 0, 1);
		normal[3] = Vector3D(-1, 0, 0);
		normal[4] = Vector3D( 0,-1, 0);
		normal[5] = Vector3D( 0, 0,-1);

		// check starting surfels
		for (i = 0; i < 6; i++) {

			if (i == -1) {
				continue;
			}

			if (Vector3D::dotProduct(normal[i], surfelNormals[i]) < 0.0f) {
				surfelNormals[i] *= -1;
			}

			surfelsToExpand.insert(startingSurfels[i]);
		}

		if (AdditionalStartingSurfelsCheckBox->isChecked()) {
			for (j = 0; j < nofPlanes; j++) {
				if (planeIndices[j] != -1) {
					surfelsToExpand.insert(planeIndices[j]);
					qDebug("on");
				}
			}
		}

		float tolerance = 0.001f * this->ToleranceValueSlider->value();

		while (surfelsToExpand.size() > 0) {

			while (surfelsToExpand.size() > 0) {
				
				index = *surfelsToExpand.begin();

				goodNormal = surfelNormals[index];
				search.setSourcePoint(surfelPositions[index]);
				search.setNofNeighbours(nofNeighbours);
				okSurfels.insert(index);

				for (int j = 1; j < nofNeighbours; j++) {
					if (okSurfels.find(search.getNeighbourPositionIndex(j)) == okSurfels.end()) {
						// set start point
						newNormal = surfelNormals[search.getNeighbourPositionIndex(j)];
						Edge *edge = new Edge(index, search.getNeighbourPositionIndex(j));
						
						// it turned out to be better not including the distance to the edge costs
						edge->cost = 1.0f - Vector3D::dotProduct(goodNormal, newNormal);
						
						heap.insert(edge, edge->cost);
					}	
				}

				surfelsToExpand.erase(surfelsToExpand.begin());
			}

			while (heap.getNofElements() > 0) {  
				
				Edge *edge = heap.getMinIndex();
				if (edge->cost < tolerance) {
					surfelsToExpand.insert(edge->index2);
					heap.removeMin();
					delete edge;
				}
				else {
					break;
				}
			}
		}

		// delete remaining elements of the heap
		while (heap.getNofElements() > 0) {  
			Edge *edge = heap.getMinIndex();
			heap.removeMin();
			delete edge;
		}

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			if (okSurfels.find(i) == okSurfels.end()) {
				scene->setSurfelFlagOn (surfel, (SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE), true, false);
			}
			else {
				scene->setSurfelFlagOn (surfel, (SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE), false, false);
			}
		}

		qDebug("Surfels deleted: %i", nofSurfels - okSurfels.size());

		scene->emitObjectModifiedSignal();
		scene->emitRenderFrameSignal();

		delete []surfelPositions;
		delete []surfelNormals;
		if (AdditionalStartingSurfelsCheckBox->isChecked()) {
			delete []planeNormals;
			delete []planeIndices;
			delete []planeMaxDist;
		}
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_RemoveBadNormals() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;

	// Get the active object
	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		scene->removeSurfels((SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE), true);
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_FlipSelectedNormals() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;
	int i;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			if (surfel->isFlagOn((SurfelInterface::Flag)(SurfelInterface::SELECTED3 | SurfelInterface::EMPHASISE))) {
				scene->setSurfelNormal(surfel, -surfel->getNormal(), false);
			}
		}
	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

	QApplication::restoreOverrideCursor ();
}


////////////////////////////////////////////////////////////
// Distance Field Calculation
////////////////////////////////////////////////////////////


void Diffuser_DialogImpl::handle_CalculationMethod(const QString &string) {

	if (string == "Tangent Plane Estimate") {
		this->distanceFieldCalcMethod = TANGENTPLANES;
	}
	else if (string == "Optimized Weighted Least Squares") {
		this->distanceFieldCalcMethod = OPT_WEIGHTED_LEASTSQUARES;
	}
	else if (string == "Weighted Least Squares") {
		this->distanceFieldCalcMethod = WEIGHTED_LEASTSQUARES;
	}
	else if (string == "Next Point") {
		this->distanceFieldCalcMethod = NEXT_POINT;
	}
}


void Diffuser_DialogImpl::handle_ResolutionChanged(const QString &string) {

	if (string == "16x16x16") {
		this->maxDepth = 4;
	}
	else if (string == "32x32x32") {
		this->maxDepth = 5;
	}
	else if (string == "64x64x64") {
		this->maxDepth = 6;
	}
	else if (string == "128x128x128") {
		this->maxDepth = 7;
	}
	else if (string == "256x256x256") {
		this->maxDepth = 8;
	}
	else if (string == "512x512x512") {
		this->maxDepth = 9;
	}
}


////////////////////////////////////////////////////////////
// Volumetric Diffusion
////////////////////////////////////////////////////////////


void Diffuser_DialogImpl::handle_ClearGrid() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	if (this->grid) {
		delete this->grid;
		this->grid = NULL;
	}
	if (this->narrowBand) {
		delete this->narrowBand;
		this->narrowBand = NULL;
	}
	if (this->particleSim) {
		delete this->particleSim;
		this->particleSim = NULL;
	}
	
	qDebug("Volumetric grid deleted");

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_CalculateGrid () {

	QApplication::setOverrideCursor (Qt::waitCursor);

	Vector3D	bBoxMin;
	Vector3D	bBoxMax;
    QTime		t;

	// Volumetric grid generation
	qDebug("Volumetric grid generation");
	t.start();
	this->getBoundingBox(bBoxMin, bBoxMax);
	
	if (bBoxMin == bBoxMax) {
		QApplication::restoreOverrideCursor ();
		return;
	}

	if (this->grid) {
		delete this->grid;
		this->grid = NULL;
	}
	if (this->narrowBand) {
		delete this->narrowBand;
		this->narrowBand = NULL;
	}
	if (this->particleSim) {
		delete this->particleSim;
		this->particleSim = NULL;
	}
	
	this->grid = new VolGrid();
	this->grid->distanceFieldCalcMethod = this->distanceFieldCalcMethod;
	this->grid->distanceFieldNofNeighbours = this->DistanceFieldNofNeighborsSpinBox->value();

	int narrowbandWidth = this->NarrowBandSpinBox->value();
	float bboxSize;
	bool isNumber;
	bboxSize = this->BBoxSizeLineEdit->text().toFloat(&isNumber);
		if (!isNumber || bboxSize <= 0.0f) {
			this->SmoothingInfluenceLineEdit->setText(QString("1.0"));
			bboxSize = 1.0f;
			QApplication::restoreOverrideCursor ();
			return;
		}

	grid->createVolGrid(bBoxMin, bBoxMax, bboxSize, this->maxDepth, narrowbandWidth);
	qDebug("");
	qDebug("Time for volumetric grid generation elapsed: %d ms", t.elapsed());
	qDebug("");

	// Statistics gatherer
//	OcTreeStatistics stats;
//	stats.collectStatistics(grid);
//  stats.reportStatistics();
//	stats.clear();

	// Narrow band calculation
	t.start();
	qDebug("");
	qDebug("Narrow band calculation");
	this->narrowBand = new VolGridNarrowBandIterator(this->grid);
	this->narrowBand->getNarrowBand();
	qDebug("Time for narrow band calculation elapsed: %d ms", t.elapsed());

	handle_YZPlane();

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_FilterChanged(const QString& string) {
		
	if (string == "3x3x3 box") {
		this->Filter0_LineEdit->setText(QChar('1'));
		this->Filter1_LineEdit->setText(QChar('1'));
		this->Filter2_LineEdit->setText(QChar('1'));
		this->Filter3_LineEdit->setText(QChar('1'));
		this->Filter4_LineEdit->setText(QChar('1'));
		this->Filter5_LineEdit->setText(QChar('1'));
		this->Filter6_LineEdit->setText(QChar('1'));
		this->Filter7_LineEdit->setText(QChar('1'));
		this->Filter8_LineEdit->setText(QChar('1'));
		this->Filter9_LineEdit->setText(QChar('1'));
		this->Filter10_LineEdit->setText(QChar('1'));
		this->Filter11_LineEdit->setText(QChar('1'));
		this->Filter12_LineEdit->setText(QChar('1'));
		this->Filter13_LineEdit->setText(QChar('1'));
		this->Filter14_LineEdit->setText(QChar('1'));
		this->Filter15_LineEdit->setText(QChar('1'));
		this->Filter16_LineEdit->setText(QChar('1'));
		this->Filter17_LineEdit->setText(QChar('1'));
		this->Filter18_LineEdit->setText(QChar('1'));
		this->Filter19_LineEdit->setText(QChar('1'));
		this->Filter20_LineEdit->setText(QChar('1'));
		this->Filter21_LineEdit->setText(QChar('1'));
		this->Filter22_LineEdit->setText(QChar('1'));
		this->Filter23_LineEdit->setText(QChar('1'));
		this->Filter24_LineEdit->setText(QChar('1'));
		this->Filter25_LineEdit->setText(QChar('1'));
		this->Filter26_LineEdit->setText(QChar('1'));
	}
	else if (string == "7-part plus") {
		this->Filter0_LineEdit->setText(QChar('0'));
		this->Filter1_LineEdit->setText(QChar('0'));
		this->Filter2_LineEdit->setText(QChar('0'));
		this->Filter3_LineEdit->setText(QChar('0'));
		this->Filter4_LineEdit->setText(QChar('1'));
		this->Filter5_LineEdit->setText(QChar('0'));
		this->Filter6_LineEdit->setText(QChar('0'));
		this->Filter7_LineEdit->setText(QChar('0'));
		this->Filter8_LineEdit->setText(QChar('0'));
		this->Filter9_LineEdit->setText(QChar('0'));
		this->Filter10_LineEdit->setText(QChar('1'));
		this->Filter11_LineEdit->setText(QChar('0'));
		this->Filter12_LineEdit->setText(QChar('1'));
		this->Filter13_LineEdit->setText(QChar('1'));
		this->Filter14_LineEdit->setText(QChar('1'));
		this->Filter15_LineEdit->setText(QChar('0'));
		this->Filter16_LineEdit->setText(QChar('1'));
		this->Filter17_LineEdit->setText(QChar('0'));
		this->Filter18_LineEdit->setText(QChar('0'));
		this->Filter19_LineEdit->setText(QChar('0'));
		this->Filter20_LineEdit->setText(QChar('0'));
		this->Filter21_LineEdit->setText(QChar('0'));
		this->Filter22_LineEdit->setText(QChar('1'));
		this->Filter23_LineEdit->setText(QChar('0'));
		this->Filter24_LineEdit->setText(QChar('0'));
		this->Filter25_LineEdit->setText(QChar('0'));
		this->Filter26_LineEdit->setText(QChar('0'));
	}
	else if (string == "Gaussian I") {
		this->Filter0_LineEdit->setText(QChar('1'));
		this->Filter1_LineEdit->setText(QChar('2'));
		this->Filter2_LineEdit->setText(QChar('1'));
		this->Filter3_LineEdit->setText(QChar('2'));
		this->Filter4_LineEdit->setText(QChar('4'));
		this->Filter5_LineEdit->setText(QChar('2'));
		this->Filter6_LineEdit->setText(QChar('1'));
		this->Filter7_LineEdit->setText(QChar('2'));
		this->Filter8_LineEdit->setText(QChar('1'));
		this->Filter9_LineEdit->setText(QChar('2'));
		this->Filter10_LineEdit->setText(QChar('4'));
		this->Filter11_LineEdit->setText(QChar('2'));
		this->Filter12_LineEdit->setText(QChar('4'));
		this->Filter13_LineEdit->setText(QChar('8'));
		this->Filter14_LineEdit->setText(QChar('4'));
		this->Filter15_LineEdit->setText(QChar('2'));
		this->Filter16_LineEdit->setText(QChar('4'));
		this->Filter17_LineEdit->setText(QChar('2'));
		this->Filter18_LineEdit->setText(QChar('1'));
		this->Filter19_LineEdit->setText(QChar('2'));
		this->Filter20_LineEdit->setText(QChar('1'));
		this->Filter21_LineEdit->setText(QChar('2'));
		this->Filter22_LineEdit->setText(QChar('4'));
		this->Filter23_LineEdit->setText(QChar('2'));
		this->Filter24_LineEdit->setText(QChar('1'));
		this->Filter25_LineEdit->setText(QChar('2'));
		this->Filter26_LineEdit->setText(QChar('1'));
	}
	else if (string == "Gaussian II") {
		this->Filter0_LineEdit->setText(QChar('1'));
		this->Filter1_LineEdit->setText(QChar('1'));
		this->Filter2_LineEdit->setText(QChar('1'));
		this->Filter3_LineEdit->setText(QChar('1'));
		this->Filter4_LineEdit->setText(QChar('2'));
		this->Filter5_LineEdit->setText(QChar('1'));
		this->Filter6_LineEdit->setText(QChar('1'));
		this->Filter7_LineEdit->setText(QChar('1'));
		this->Filter8_LineEdit->setText(QChar('1'));
		this->Filter9_LineEdit->setText(QChar('1'));
		this->Filter10_LineEdit->setText(QChar('2'));
		this->Filter11_LineEdit->setText(QChar('1'));
		this->Filter12_LineEdit->setText(QChar('2'));
		this->Filter13_LineEdit->setText(QChar('4'));
		this->Filter14_LineEdit->setText(QChar('2'));
		this->Filter15_LineEdit->setText(QChar('1'));
		this->Filter16_LineEdit->setText(QChar('2'));
		this->Filter17_LineEdit->setText(QChar('1'));
		this->Filter18_LineEdit->setText(QChar('1'));
		this->Filter19_LineEdit->setText(QChar('1'));
		this->Filter20_LineEdit->setText(QChar('1'));
		this->Filter21_LineEdit->setText(QChar('1'));
		this->Filter22_LineEdit->setText(QChar('2'));
		this->Filter23_LineEdit->setText(QChar('1'));
		this->Filter24_LineEdit->setText(QChar('1'));
		this->Filter25_LineEdit->setText(QChar('1'));
		this->Filter26_LineEdit->setText(QChar('1'));		
	}
}


void Diffuser_DialogImpl::getFilterValues(VolGrid *grid) {

	bool ok;
	int value;
	QString valueString;

	valueString = this->Filter0_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[0] = value;
	}
	else {
		grid->filter[0] = 1;
	}

	valueString = this->Filter1_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[1] = value;
	}
	else {
		grid->filter[1] = 1;
	}

	valueString = this->Filter2_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[2] = value;
	}
	else {
		grid->filter[2] = 1;
	}

	valueString = this->Filter3_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[3] = value;
	}
	else {
		grid->filter[3] = 1;
	}

	valueString = this->Filter4_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[4] = value;
	}
	else {
		grid->filter[4] = 1;
	}

	valueString = this->Filter5_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[5] = value;
	}
	else {
		grid->filter[5] = 1;
	}

	valueString = this->Filter6_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[6] = value;
	}
	else {
		grid->filter[6] = 1;
	}

	valueString = this->Filter7_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[7] = value;
	}
	else {
		grid->filter[7] = 1;
	}

	valueString = this->Filter8_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[8] = value;
	}
	else {
		grid->filter[8] = 1;
	}

	valueString = this->Filter9_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[9] = value;
	}
	else {
		grid->filter[9] = 1;
	}

	valueString = this->Filter10_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[10] = value;
	}
	else {
		grid->filter[10] = 1;
	}

	valueString = this->Filter11_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[11] = value;
	}
	else {
		grid->filter[11] = 1;
	}

	valueString = this->Filter12_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[12] = value;
	}
	else {
		grid->filter[12] = 1;
	}

	valueString = this->Filter13_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[13] = value;
	}
	else {
		grid->filter[13] = 1;
	}

	valueString = this->Filter14_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[14] = value;
	}
	else {
		grid->filter[14] = 1;
	}

	valueString = this->Filter15_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[15] = value;
	}
	else {
		grid->filter[15] = 1;
	}

	valueString = this->Filter16_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[16] = value;
	}
	else {
		grid->filter[16] = 1;
	}

	valueString = this->Filter17_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[17] = value;
	}
	else {
		grid->filter[17] = 1;
	}

	valueString = this->Filter18_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[18] = value;
	}
	else {
		grid->filter[18] = 1;
	}

	valueString = this->Filter19_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[19] = value;
	}
	else {
		grid->filter[19] = 1;
	}

	valueString = this->Filter20_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[20] = value;
	}
	else {
		grid->filter[20] = 1;
	}

	valueString = this->Filter21_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[21] = value;
	}
	else {
		grid->filter[21] = 1;
	}

	valueString = this->Filter22_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[22] = value;
	}
	else {
		grid->filter[22] = 1;
	}

	valueString = this->Filter23_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[23] = value;
	}
	else {
		grid->filter[23] = 1;
	}

	valueString = this->Filter24_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[24] = value;
	}
	else {
		grid->filter[24] = 1;
	}

	valueString = this->Filter25_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[25] = value;
	}
	else {
		grid->filter[25] = 1;
	}

	valueString = this->Filter26_LineEdit->text();
	value = valueString.toInt(&ok);
	if (ok) {
		grid->filter[26] = value;
	}
	else {
		grid->filter[26] = 1;
	}

}


void Diffuser_DialogImpl::handle_Smoother() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	float alpha;
	bool isNumber; 
	QTime t;

	if (this->grid != NULL && this->narrowBand != NULL) {	

		t.start();

		// get filter values
		getFilterValues(this->grid);

		// validate input
		alpha = this->SmoothingInfluenceLineEdit->text().toFloat(&isNumber);
		if (!isNumber || alpha > 1.0f || alpha <= 0.0f) {
			this->SmoothingInfluenceLineEdit->setText(QString("1.0"));
			alpha = 1.0;
			QApplication::restoreOverrideCursor ();
			return;
		}

		qDebug("");
		qDebug("Smoothing ...");
		int nOfIterations = this->SmoothingNumberOfIterationsSpinBox->value();
		this->narrowBand->smooth(nOfIterations, alpha);
		qDebug("Time for volumetric smoothing elapsed: %d ms", t.elapsed());
		qDebug("");

		updateSlicePicture(this->DistanceFieldSlider->value(), this->plane);
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_ExpandOctree() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);
	QTime t;
	t.start();
	
	if (this->grid != NULL && this->narrowBand != NULL) {	
	
		qDebug("");
		qDebug("Expand OcTree");
		this->narrowBand->growVolGrid(LargestHoleSpinBox->value());

		PLANE plane;
		if (this->XYPlaneRadioButton->isChecked()) {
			plane = XY_PLANE;
		}
		else if (this->YZPlaneRadioButton->isChecked()) {
			plane = YZ_PLANE;
		}
		else {
			plane = XZ_PLANE;
		}
		qDebug("");
		qDebug("Time for Octree expanding elapsed: %d ms", t.elapsed());
		
		updateSlicePicture(this->DistanceFieldSlider->value(), plane);
	}
	
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_FindBoundaryNodes() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	QTime t;
	t.start();

	if (this->grid != NULL && this->narrowBand != NULL) {		
		qDebug("");
		qDebug("Find boundary nodes");
		this->narrowBand->findBoundaryNodes(this->SignChangeSpinBox->value(), this->InvalidVoxelsSpinBox->value());
		qDebug("");	
		qDebug("Time for boundary detection elapsed: %d ms", t.elapsed());
		
		updateSlicePicture(this->DistanceFieldSlider->value(), this->plane);
	}
	
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_ClearBoundaryNodes() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	if (this->grid != NULL && this->narrowBand != NULL) {
		this->narrowBand->clearBoundaryNodes();
	}
	
	updateSlicePicture(this->DistanceFieldSlider->value(), this->plane);

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_FillHoles() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	QTime t;
	bool isNumber;

	if (this->grid != NULL && this->narrowBand != NULL) {		
	
		t.start();
		qDebug("");
		qDebug("Hole Filling ...");
	
		int nOfIterations = this->NumberOfIterationsSpinBox->value();
		bool blending = this->BlendingCheckBox->isChecked();
		
		float blendingConstant = this->BlendingLineEdit->text().toFloat(&isNumber);
		if (!isNumber || blendingConstant > 1.0f || blendingConstant <= 0.0f) {
			this->BlendingLineEdit->setText(QString("1.0"));
			blendingConstant = 1.0;
			QApplication::restoreOverrideCursor ();
			return;
		}
		
		// get filter values
		getFilterValues(this->grid);

		this->narrowBand->diffuse(nOfIterations, blending, blendingConstant);
		
		qDebug("");	
		qDebug("Time for hole filling elapsed: %d ms", t.elapsed());
		updateSlicePicture(this->DistanceFieldSlider->value(), this->plane);
	}
	
	QApplication::restoreOverrideCursor ();
}


////////////////////////////////////////////////////////////
// Surface Reconstruction
////////////////////////////////////////////////////////////


void Diffuser_DialogImpl::handle_GenerateSurfelSurface() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);
	bool isNumber;

	// ellipcitcal marching cubes
	if (this->grid != NULL) {
		qDebug("");
		qDebug("Elliptical Marching Cubes");
		VolGridEllipticalMarchingCubes *surfelSurface = new VolGridEllipticalMarchingCubes(this->grid);
		float isoValue = this->IsoValueLineEdit->text().toFloat (&isNumber);
		
		if (!isNumber) {
			isoValue = 0.0f;
			this->IsoValueLineEdit->setText(QString("0.0"));
		}

		surfelSurface->build(isoValue);
		qDebug("");
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_GenerateMesh() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	if (this->grid != NULL) {
		qDebug("");
		qDebug("Mesh generation");

	    QString fn = QFileDialog::getSaveFileName("Mesh.raw", "Triangle List (*.raw)", this);

		if (!fn.isEmpty()) {
			// assume that name does not include .raw
			if (!fn.contains(".raw")) {
				fn.append(".raw");
			}
			VolGridMesh *mesh = new VolGridMesh(this->grid, fn, this);
			mesh->build();
			delete mesh;
			qDebug("");
		}
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_InitParticleSimulation() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	
	if (this->grid != NULL) {
		qDebug("");
		qDebug("Particle Simulation");
		VolGridParticleSimulation *particleSim = new VolGridParticleSimulation(this->grid, this->PSTargetSizeSpinBox->value());
		this->particleSim = particleSim;
		this->particleSim->variationInfluence = (float)this->PSVariationInfluenceSlider->value() / (this->PSVariationInfluenceSlider->maxValue());
		particleSim->build();
		qDebug("");
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_Repel() {

	QApplication::setOverrideCursor (Qt::waitCursor);

	bool isNumber;

	if (this->particleSim != NULL) {

		this->particleSim->numNeighbours = this->PSNumberOfNeighborsSpinBox->value();
		this->particleSim->numIterations = this->PSNumberOfIterationsSpinBox->value();
		this->particleSim->damping = this->PSDampingLineEdit->text().toFloat (&isNumber);
		if (!isNumber) {
			this->PSDampingLineEdit->setText(QString("1.0"));
			this->particleSim->damping = 1.0f;
		}
		this->particleSim->forceScale = this->PSForceScaleLineEdit->text().toFloat (&isNumber);
		if (!isNumber) {
			this->PSForceScaleLineEdit->setText(QString("1.0"));
			this->particleSim->forceScale = 1.0f;
		}

		particleSim->repel();
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_MLS() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	int nofNeighbours = this->MLSnofNeighboursSpinBox->value();
	bool approx = MLSApproximationCheckBox->isChecked();
	bool recalcNormals = MLSRecalculateNormalsCheckBox->isChecked();
	bool adjustNormals = MLSAdjustNormalsCheckBox->isChecked();
	
	MLSApplication *mls = new MLSApplication(nofNeighbours, this->polynomial, approx, recalcNormals, adjustNormals);
	
	mls->init();
	mls->run();

	delete mls;
	
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_PolynomialChanged(const QString &string) {

	if (string == "3rd Degree Polynomial") {
		this->polynomial = 3;
	}
	else if (string == "4th Degree Polynomial") {
		this->polynomial = 4;
	}
}


////////////////////////////////////////////////////////////
// General
////////////////////////////////////////////////////////////


void Diffuser_DialogImpl::getBoundingBox(Vector3D &minimum, Vector3D &maximum) {

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface	 *surfel;
	Vector3D		 point;

	maximum = minimum = Vector3D(0.0f, 0.0f, 0.0f);
	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {

		surfels = object->getSurfelCollection();
		if (surfels->getNofSurfels() == 0) {
			return;
		}
		// init values for the bounding box
		maximum = minimum = surfels->getFirstSurfel()->getPosition();

		for (surfel = surfels->getFirstSurfel(); surfel != 0; surfel = surfels->getNextSurfel()) {
			point = surfel->getPosition();
			maximum[0] = MAX(maximum[0], point[0]);
			maximum[1] = MAX(maximum[1], point[1]);
			maximum[2] = MAX(maximum[2], point[2]);
			minimum[0] = MIN(minimum[0], point[0]);
			minimum[1] = MIN(minimum[1], point[1]);
			minimum[2] = MIN(minimum[2], point[2]);
		}
	}
}


void Diffuser_DialogImpl::handle_XYPlane() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	this->plane = XY_PLANE;
	this->YZPlaneRadioButton->setChecked(false);
	this->XZPlaneRadioButton->setChecked(false);
	this->XYPlaneRadioButton->setChecked(true);
	if (this->grid) {
		this->DistanceFieldSlider->setMaxValue(this->grid->getSliceDimension(this->grid->getMaxDepth()) - 1);
		this->DistanceFieldSlider->setValue(0);
		this->updateSlicePicture(this->DistanceFieldSlider->value(), XY_PLANE);
	}
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_YZPlane() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	this->plane = YZ_PLANE;
	this->XZPlaneRadioButton->setChecked(false);
	this->XYPlaneRadioButton->setChecked(false);
	this->YZPlaneRadioButton->setChecked(true);
	if (this->grid) {
		this->DistanceFieldSlider->setMaxValue(this->grid->getSliceDimension(this->grid->getMaxDepth()) - 1);
		this->DistanceFieldSlider->setValue(0);
		this->updateSlicePicture(this->DistanceFieldSlider->value(), YZ_PLANE);
	}
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_XZPlane() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	this->plane = XZ_PLANE;
	this->YZPlaneRadioButton->setChecked(false);
	this->XYPlaneRadioButton->setChecked(false);
	this->XZPlaneRadioButton->setChecked(true);
	if (this->grid) {
		this->DistanceFieldSlider->setMaxValue(this->grid->getSliceDimension(this->grid->getMaxDepth()) - 1);
		this->DistanceFieldSlider->setValue(0);
		this->updateSlicePicture(this->DistanceFieldSlider->value(), XZ_PLANE);
	}
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_SliderValueChanged (int value) {
	this->updateSlicePicture(value, this->plane);
}


void Diffuser_DialogImpl::updateSlicePicture(int sliceNumber, PLANE plane) {
	
	QPixmap *oldPixmap;
	int slicerValue;

	if (this->grid) {
		slicerValue = this->DistanceFieldSlider->value();
		oldPixmap = this->actualPixmap;
		this->actualPixmap = this->createSliceImage(this->grid, plane, slicerValue, this->grid->getMaxDepth());
		this->DistanceFieldPixmapLabel->setPixmap(*this->actualPixmap);
		if (oldPixmap != NULL)
			delete oldPixmap;
	}
}


void Diffuser_DialogImpl::handle_DisableHoleMarker() {

	QApplication::setOverrideCursor (Qt::waitCursor);
	updateSlicePicture(this->DistanceFieldSlider->value(), this->plane);	
	QApplication::restoreOverrideCursor ();
}


QPixmap* Diffuser_DialogImpl::createSliceImage(VolGrid *grid, PLANE plane, int sliceNumber, int level) {

	QRgb color;
	int x, y, val;
	int dim;
	float *values;
	
	float *boundary;

	dim = pow(2, level);
	QImage *sliceImage = new QImage(dim, dim, 32);
	QPixmap *slicePixmap = new QPixmap();

	if (grid->getSlice(level, sliceNumber, plane)) {

		values = grid->getSlice();
		boundary = grid->boundary;

		for (x = 0; x < dim; x++) {
			for (y = 0; y < dim; y++, values++, boundary++) {
				val = (int ) (*values);
				// grey is the iso surface, white is inside, black is outside
				if (abs(val) <= (0.5 * grid->getNarrowBandWidth())) {
					val = 127 +  (-1) * ((int)((val * 255)/(grid->getNarrowBandWidth())));
					color = qRgb (val,val,val);
				}
				else {
					color = qRgb (0, 0, 255);
				}
				
				if (!this->DisableHoleMarkerCheckBox->isChecked()) {
					if (*boundary == 1.0f) {
						color = qRgb (255, 0, 0);
					}
					if (*boundary == 2.0f) {
						color = qRgb (220, 0, 0);
					}
				}
				
				sliceImage->setPixel (x, y, color);				
			}
		}

		slicePixmap->convertFromImage(*sliceImage);
	}

	delete sliceImage;
	return slicePixmap;
}


////////////////////////////////////////////////////////////
// Debug
////////////////////////////////////////////////////////////


void Diffuser_DialogImpl::handle_LoadGrid() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);
    QTime		t;
	t.start();

	if (this->narrowBand) {
		delete this->narrowBand;
		this->narrowBand = NULL;
	}

	if (this->grid != NULL) {
		delete this->grid;
		this->grid = NULL;
	}
	
	VolGrid *grid = new VolGrid();
	this->grid = grid;

	this->reader = new VolGridReader();
	this->reader->grid = this->grid;
	char *filename = "VolGrid.grd";
	this->reader->fp = fopen(filename, "rb");
	this->reader->load(this->reader->fp);

	this->grid->slice = new float[int(pow(pow(2, this->grid->maxDepth), 2))];
	this->grid->boundary = new float[int(pow(pow(2, this->grid->maxDepth), 2))];	

	// precalculate length table for faster access
   	this->grid->lengthTable[0] = 2 * this->grid->length;
  	for (int i = 1; i < 16; i++)
  		this->grid->lengthTable[i] = 0.5 * this->grid->lengthTable[i - 1];

  	this->grid->nodeFactory = new OcTreeTraversalNodePool<VolGridTraversalNode>(32, 1);
	
	this->narrowBand = new VolGridNarrowBandIterator(this->grid);
	this->narrowBand->getNarrowBand();

	this->plane = YZ_PLANE;
	this->XZPlaneRadioButton->setChecked(false);
	this->XYPlaneRadioButton->setChecked(false);
	this->YZPlaneRadioButton->setChecked(true);
	if (this->grid != NULL) {
		this->DistanceFieldSlider->setMaxValue(this->grid->getSliceDimension(this->grid->getMaxDepth()) - 1);
		this->DistanceFieldSlider->setValue(0);
		this->updateSlicePicture(this->DistanceFieldSlider->value(), YZ_PLANE);
	}
	
	qDebug("Time elapsed for loading volumetric grid: %d ms", t.elapsed());
	qDebug("");
	// Statistics gatherer
	OcTreeStatistics stats;
	stats.collectStatistics(grid);
    stats.reportStatistics();
	stats.clear();

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_SaveGrid() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	if (!this->reader) {
		this->reader = new VolGridReader();
	}
	
	this->reader->grid = this->grid;
	char *filename = "VolGrid.grd";
	this->reader->fp = fopen(filename, "wb");
	this->reader->save(this->reader->fp);

	delete this->reader;
	this->reader = NULL;
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_WriteFile() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);
	

	Scene            *scene;
	Object           *object;
	QString			 name;
	SurfelCollection *surfels;
	SurfelInterface	 *surfel;
	Vector3D		 point;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		name = object->getName();
	}

	QString fn = QFileDialog::getSaveFileName(name, "Points with Normals (*.pwn)", this);

    if (!fn.isEmpty()) {
		// assume that name does not include .pwn
		if (!fn.contains(".pwn")) {
			fn.append(".pwn");
		}
		
		QFile fp(fn);
		if (!fp.open( IO_WriteOnly)) {
			QMessageBox::warning(this,"I/O Error", QString("The file could not be opened for writing.\n\n") + fn);
			QApplication::restoreOverrideCursor ();
			return;
		}

		if (object) {
			QTextStream ts(&fp);
			surfels = object->getSurfelCollection();
			// init values for the bounding box

			ts << scene->getNofSurfels() << "\r\n";

			float factor = 1.0f;

			for (surfel = surfels->getFirstSurfel(); surfel != 0; surfel = surfels->getNextSurfel()) {
				Vector3D pos = surfel->getPosition();
				ts << pos[0]*factor << " " << pos[1]*factor << " " << pos[2]*factor << "\r\n";
			}


			for (surfel = surfels->getFirstSurfel(); surfel != 0; surfel = surfels->getNextSurfel()) {
				Vector3D norm = surfel->getNormal();
				ts << norm[0] << " " << norm[1] << " " << norm[2] << "\r\n";
			}
		}	
	
		fp.flush();
		fp.close();
	}

	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_ReadFile() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *newObject;
	SurfelCollection *surfels;
	SurfelInterface	 *surfel;
	Vector3D		 point;
	QString			 str;

	scene = Scene::getInstance();

	newObject = new Object(ObjectNameLineEdit->text().latin1());

	QString fn = QFileDialog::getOpenFileName(QString::null, "Points with Normals (*.pwn)", this);

	if (!fn.isEmpty() && newObject) {

		QFile fp(fn);
		if (!fp.open( IO_ReadOnly)) {
			QMessageBox::warning(this,"I/O Error", QString("The file could not be opened for reading.\n\n") + fn);
			QApplication::restoreOverrideCursor ();
			return;
		}
		
		QTextStream ts(&fp);

		int nofSurfels = 0;
		surfels = newObject->getSurfelCollection();
		SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::RADIUS | SurfelInterface::POSITION;
		surfels->setPropertyDescriptor(pd);
		
		str = ts.readLine();
		sscanf(str,"%i\n", &nofSurfels);
		surfels->reserve(nofSurfels);

		for (int i = 0; i < nofSurfels; i++) {
			surfel = surfels->addSurfel(false);
			Vector3D pos;
			str = ts.readLine();
			sscanf(str,"%f %f %f\n", &pos[0], &pos[1], &pos[2]);
//			pos *= 250.0f;
			scene->setSurfelPosition (surfel, pos , false);	
		}

		for (surfel = surfels->getFirstSurfel(); surfel != 0; surfel = surfels->getNextSurfel()) {
			Vector3D norm;
			str = ts.readLine();
			sscanf(str,"%f %f %f\n", &norm[0], &norm[1], &norm[2]);
			scene->setSurfelNormal (surfel, norm , false);
			scene->setSurfelRadius(surfel, 1.0, false);
		}

		scene->addObject(newObject);	

		fp.flush();
		fp.close();
	}
	
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	QApplication::restoreOverrideCursor ();
}


void Diffuser_DialogImpl::handle_WriteStatistics() {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	Scene            *scene;
	Object           *object;
	QString			 str;
	QString			 fn;

	scene = Scene::getInstance();
	object = scene->getActiveObject();

	fn = QFileDialog::getSaveFileName("Stats", "Statistical Data (*.txt)", this);

    if (!fn.isEmpty()) {
		// assume that name does not include .txt
		if (!fn.contains(".txt")) {
			fn.append(".txt");
		}

		QFile fp(fn);
		if (!fp.open( IO_WriteOnly)) {
			QMessageBox::warning(this,"I/O Error", QString("The file could not be opened for writing.\n\n") + fn);
			QApplication::restoreOverrideCursor ();
			return;
		}
		
		QTextStream ts(&fp);
		
		ts << "Surface Reconstruction Stats:" << "\r\n";
		ts << "=============================" << "\r\n";
		
		if (object) {
			ts << "\r\n" << "Input Model:" << "\r\n";
			ts << "============" << "\r\n";
			ts << "Object Name: " << object->getName().ascii() << "\r\n";
			ts << "Number of Surfels: " << object->getSurfelCollection()->getNofSurfels() << "\r\n";
		}

		ts << "\r\n" << "Preprocessing:" << "\r\n";
		ts << "==============" << "\r\n";
		ts << "\r\n" << "\tBilateral Smoothing: " << "\r\n";
		ts << "\tSmooth Normals: " << "\r\n";
		ts << "\tCenter of Mass: " << "\r\n";
		ts << "\tFlipped Normal Detection: " << "\r\n";

		ts << "\r\nDistance Field:" << "\r\n";
		ts << "===============" << "\r\n";
		ts << "\r\n" << "\tCalculation Method: " << this->CalculationMethodComboBox->currentText().ascii() << "\r\n";
		ts << "\tResolution: " << this->ResolutionComboBox->currentText().ascii() << "\r\n";
		ts << "\tBounding Box Size: " << this->BBoxSizeLineEdit->text().ascii() << "\r\n";
		ts << "\tNarrow Band (Number of Voxels): " << this->NarrowBandSpinBox->value() << "\r\n";
		ts << "\tNumber of Neighbors: " << this->DistanceFieldNofNeighborsSpinBox->value() << "\r\n";
		
		ts << "\r\n" << "Volumetric Diffusion:" << "\r\n";
		ts << "=====================" << "\r\n";
		ts << "\r\n" << "\tFilter:" << "\r\n";
		ts << "\t-------" << "\r\n";
		ts << "\tType: " << this->FilterComboBox->currentText().ascii() << "\r\n";
		ts << "\r\n" << "\tSmoothing:" << "\r\n";
		ts << "\t----------" << "\r\n";
		ts << "\tSmoothing Influence: " << this->SmoothingInfluenceLineEdit->text().ascii() << "\r\n";
		ts << "\tSmoothing Iterations: " << this->SmoothingNumberOfIterationsSpinBox->value() << "\r\n";
		
		ts << "\r\n" << "\tHole Filling:" << "\r\n";
		ts << "\t-------------" << "\r\n";
		ts << "\tNumber of Invalid Voxels: " << this->InvalidVoxelsSpinBox->value() << "\r\n";
		ts << "\tNumber of Voxels with Sign Change: " << this->SignChangeSpinBox->value() << "\r\n";
		ts << "\tLargest Hole (Number of Voxels): " << this->LargestHoleSpinBox->value() << "\r\n";
		ts << "\tNumber of Iterations: " << this->NumberOfIterationsSpinBox->value() << "\r\n";
		if (BlendingCheckBox->isChecked()) {
			ts << "\tPerform Blending: Checked" << "\r\n";
		}
		else {
			ts << "\tPerform Blending: Not checked" << "\r\n";
		}
		ts << "\tBlending Constant: " << this->BlendingLineEdit->text().ascii() << "\r\n";

		ts << "\r\nSurface Reconstruction:" << "\r\n";
		ts << "=======================" << "\r\n";
		ts << "\r\n" << "\tMethod: Elliptical Marching Cubes" << "\r\n";
		ts << "\t---------------------------------" << "\r\n";
		ts << "\tOffsetting: " << this->IsoValueLineEdit->text().ascii() << "\r\n";
		ts << "\r\n" << "\tMethod: Particle Simulation" << "\r\n";
		ts << "\t---------------------------" << "\r\n";
		ts << "\tTarget Size: " << this->PSTargetSizeSpinBox->value() << "\r\n";
		ts << "\tVariation Influence: " << (float) (this->PSVariationInfluenceSlider->value() / 1000) << "\r\n";
		ts << "\tNumber of Neighbors: " << this->PSNumberOfNeighborsSpinBox->value() << "\r\n";
		ts << "\tIterations: " << this->PSNumberOfIterationsSpinBox->value() << "\r\n";
		ts << "\tForce Scale: " << this->PSForceScaleLineEdit->text().ascii() << "\r\n";
		ts << "\tDamping: " << this->PSDampingLineEdit->text().ascii() << "\r\n";
		ts << "\r\n" << "\tMoving Least Squares:" << "\r\n";
		ts << "\t---------------------" << "\r\n";
		ts << "\tNumber of Neighbors: " << this->MLSnofNeighboursSpinBox->value() << "\r\n";
		if (this->MLSRecalculateNormalsCheckBox->isChecked()) { 
			ts << "\tRecalculate Normals: Checked" << "\r\n";
		}
		if (this->MLSApproximationCheckBox->isChecked()) { 
			ts << "\tApproximation: Checked" << "\n";
		}
		if (this->MLSAdjustNormalsCheckBox->isChecked()) { 
			ts << "\tAdjust Normals: Checked" << "\r\n";
		}
		ts << "\tPolynomial: " << this->MLSPolynomialFilterComboBox->currentText().ascii() << "\r\n";

		fp.flush();
		fp.close();
	}
	
	QApplication::restoreOverrideCursor ();
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
