// Title:   MLSApplication.h
// Created: Thu Sep 25 15:11:20 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __MLSAPPLICATION_H_
#define __MLSAPPLICATION_H_

#include "VolGrid.h"
#include "PlaneFit.h"

/**
 * This is a application of the moving least squares projection procedure
 */
class MLSApplication {

public:
   /**
	* @param numberOfNeighbours
	*			number of neighbours for <i>Moving Least Squares</i> projection.
	* @param polynomial
	*			degree of polynomial
	* @param approximation
	*			simplification of the <i>Moving Least Squares</i> projection
	* @param recalculateNormals
	*			use weighted least squares covariance matrix to recalculate normals
	* @param adjustNormals
	*			use new normal according to the polynomial approximation
	*/
	MLSApplication(int numberOfNeighbours, int polynomial, bool approximation, bool recalculateNormals, bool adjustNormals) {
		this->numberOfNeighbours = numberOfNeighbours;
		if (polynomial == 4) {
			this->polynomial = 4;
		}
		else {
			this->polynomial = 3;
		}
		this->approximation = approximation;
		this->recalculateNormals = recalculateNormals;
		this->adjustNormals = adjustNormals;
		this->mls = NULL;
		this->positions = NULL;
		this->newPositions = NULL;
		this->normals = NULL;
		this->newNormals = NULL;
		this->nbHood = NULL;
		
	};

	~MLSApplication() {
		if (mls) {
			delete mls;
			this->mls = NULL;
		}
		if (positions) {
			delete []positions;
			this->positions = NULL;
		}
		if (newPositions) {
			delete []newPositions;
			this->newPositions = NULL;
		}
		if (normals) {
			delete []normals;
			this->normals = NULL;
		}
		if (newNormals) {
			delete []newNormals;
			this->newNormals = NULL;
		}
		if (nbHood) {
			delete nbHood;
			this->nbHood = NULL;
		}
	};
	
   /**
	* Perform <i>Moving Least Squares</i> projection.
	*/
	void run();

   /**
	* Inits the needed data structure
	*/
	void init();

public:

	// data structures
	Scene				*scene;
	Object				*object;
	Vector3D			*positions,
						*normals,
						*newPositions,
						*newNormals;
	SurfelCollection	*surfels;
	SurfelInterface		*surfel;

	/// <i>Moving Least Squares</i> Solver
	MovingLeastSquares	*mls;
	/// NeighbourHood
	NeighbourHood		*nbHood;
	/// Number of surfels
	int					numberOfSurfels;
	/// Number of Neighbours
	int					numberOfNeighbours;
	/// Degree of polynomial
	int					polynomial;
	/// Simplified moving least squares projection
	bool				approximation;
	/// Recalculate Normals
	bool				recalculateNormals;
	/// Adjust normals of the polygon
	bool				adjustNormals;
	/// False if not enough surfels
	bool				valid;
};

#endif	// __MLSAPPLICATION_H_
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
