// Title:   MovingLeastSquares.h
// Created: Thu Sep 25 15:11:13 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __MOVINGLEASTSQUARES_H_
#define __MOVINGLEASTSQUARES_H_

// Parts of the code from Numerical Recipes in C, 2nd Edition. (see Chapter 10.1 - 10.6)
// Note: Conjugate Gradient Methods not evaluated yet
//
// author Sascha Scandella (code only documented marginally)
// For details refer to "Computing and Rendering Point Set Surfaces".

#include "PlaneFit.h"
#include "../../../DataTypes/src/Vector3D.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include <math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_eigen.h>

#define ITMAX 100
#define ITMAX2 200		// here ITMAX is the maximum allowed number of iterations
#define ZEPS 1.0e-10	// EPS is a small number to rectify the special case of converging to exactly zero function value.
#define EPS 1.0e-10
#define TOL 2.0e-4f 		// Tolerance passed to dbrent.
#define MOV3(a,b,c, d,e,f) (a)=(d);(b)=(e);(c)=(f);
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
static double maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ? (maxarg1) : (maxarg2))

/**
 * This is a moving least squares implementation
 */
class MovingLeastSquares {

public:

   /**
	* @param nbHood
	*			kd-Tree for neighbour queries
	* @param numberOfSurfels
	*			number of total surfels
	* @param surfelPositions
	*			original surfel positions
	* @param surfelNormals
	*			original surfel normals
	* @param newSurfelPositions
	*			new surfel positions
	* @param newSurfelNormals
	*			new surfel normals
	* @param numberOfNeighbours
	*			number of neighbours for reference plane and projection
	* @param polynomial
	*			degree of polynomial
	* @param approximation
	*			simplification of the <i>Moving Least Squares</i> projection
	* @param recalculateNormals
	*			use weighted least squares covariance matrix to recalculate normals
	* @param adjustNormals
	*			use new normal according to the polynomial approximation
	*/
	MovingLeastSquares(NeighbourHood *nbHood, int numberOfSurfels, Vector3D *surfelPositions, Vector3D *surfelNormals,
					   Vector3D *newSurfelPositions, Vector3D *newSurfelNormals, int numberOfNeighbours,
					   int polynomial, bool approximation, bool recalculateNormals, bool adjustNormals) {


		this->nbHood = nbHood;
		this->numberOfSurfels = numberOfSurfels;
		this->positions = surfelPositions;
		this->normals = surfelNormals;
		this->newPositions = newSurfelPositions;
		this->newNormals = newSurfelNormals;
		this->numberOfNeighbours = numberOfNeighbours;
		this->polynomial = polynomial;
		this->approximation = approximation;
		this->recalculateNormals = recalculateNormals;
		this->adjustNormals = adjustNormals;
		this->normal = Vector3D(0.0f, 0.0f, 0.0f);

	};

	MovingLeastSquares() {
		
		this->nbHood = NULL;
		this->numberOfSurfels = NULL;
		this->positions = NULL;
		this->normals = NULL;

	};

	~MovingLeastSquares() {};

	// different functions needed for brent and conjugate gradient
	float evaluateFunction(float h);
	float evaluateFunction(Vector3D q);
	float evaluateDerivative(float h);
	Vector3D  evaluateDerivative(Vector3D q);
	float evaluateFunction1dim(float x);
	float evaluateDerivative1dim(float x);


   /**
	* "Numerical Recipes in C" - Routine
	*
	* Given a function f and its derivative function df, and given a bracketing triplet of abscissas ax,
	* bx, cx [such that bx is between ax and cx, and f(bx) is less than both f(ax) and f(cx)],
	* this routine isolates the minimum to a fractional precision of about tol using a modification of
	* Brent's method that uses derivatives. The abscissa of the minimum is returned as xmin, and
	* the minimum function value is returned as dbrent, the returned function value.
	*/
	float minimizeBrent(float ax, float bx, float cx, float (MovingLeastSquares::*f)(float), float (MovingLeastSquares::*df)(float), float tol, float *xmin);


   /**
	* "Numerical Recipes in C" - Routine
	*    
	* Given a function f, and given distinct initial points ax and bx, this routine searches in
	* the downhill direction (defined by the function as evaluated at the initial points) and returns
	* new points ax, bx, cx that bracket a minimum of the function. Also returned are the function
	* values at the three points, fa, fb, and fc.
	*/
	void bracketingMinimun(float *ax, float *bx, float *cx, float *fa, float *fb,float *fc, float (MovingLeastSquares::*f)(float));

   /**
	* "Numerical Recipes in C" - Routine
	*
	* Given a starting point p, Fletcher-Reeves-Polak-Ribiere minimization is performed on a
	* function f, using its gradient as calculated by a routine df. The convergence tolerance
	* on the function value is input as ftol. Returned quantities are p (the location of the minimum),
	* iter (the number of iterations that were performed), and fret (the minimum value of the
	* function). The routine dlinmin is called to perform line minimizations.
	*/
	void conjugateGradient(Vector3D p, float ftol, int *iter, float *fret, float (MovingLeastSquares::*f)(Vector3D), Vector3D (MovingLeastSquares::*df)(Vector3D));

   /**
	* "Numerical Recipes in C" - Routine
	*
	* Routine for line minimizations.
	*/
	void dlinmin(Vector3D p, Vector3D xi, float *fret, float (MovingLeastSquares::*f)(Vector3D), Vector3D (MovingLeastSquares::*df)(Vector3D));


   /** 
	* Solves the non-linear problem of the reference plane.
	*/
	void solve(Vector3D sourcePoint);

   /** 
	* Projects points on a bivariate polynomial.
	*/
	void projection();

public:

	NeighbourHood		*nbHood;
	int					numberOfNeighbours;
	int					numberOfSurfels;
	Vector3D			*positions,
						*normals,
						*newPositions,
						*newNormals;
	int					polynomial;
	bool				approximation;
	bool				adjustNormals;
	bool				recalculateNormals;
	
	Vector3D			sourcePoint;
	Vector3D			normal;
	float				_h;
	float				_t;
	
	
	Vector3D _p;
	Vector3D _xi;

protected:
	// Function pointers
	float (MovingLeastSquares::*_f)(Vector3D);
	Vector3D (MovingLeastSquares::*_df)(Vector3D);
};

#endif	// __MOVINGLEASTSQUARES_H_
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
