// Title:   OcTree.h
// Created: Thu Sep 25 15:10:39 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __OCTREE_H_
#define __OCTREE_H_

#include "myDefines.h"
#include "OcTreeTraversal.h"
#include "Pool.h"
#include "../../../DataTypes/src/Vector3D.h"

// Forward declarations
class OcTreeNode;
class OcTreeTraversalNode;
class OcTreeTraversalNodeFactory;
class OcTreeIterator;


/**
 * This is the OcTree base class.
 */
class OcTree {

public:
	
	/**
	 * Constructor calls the init method.
	 */
	OcTree() { init(MAX_LEVEL); }
    
	/**
	 * Destructor
	 */
	~OcTree() {}

	/**
	 * This method has to be called in order to use the locational codes.
	 * @param maxNumberOfLevels
	 *			specifies the maximal number of levels of an octree.
	 */
	void init(int maxNumberOfLevels);

	/**
	 * This method has to be called in order to use the locational codes.
	 * @param point
	 *			has to be element of [0,1] x [0,1] x [0.1].
	 * @param travNode
	 *			the <code>OcTreeNode</code> is stored to travNode.
	 * @return
	 *			returns true if such a octree cell exists.
	 */
	virtual bool getCellForPoint(const Vector3D &point, OcTreeTraversalNode *travNode);

	/**
	 * Determines whether a point is element of [0,1] x [0,1] x [0.1].
	 * @param point
	 *			is the point to test for.
	 * @return Returns true if the point is inside.
	 */
	bool isPointInside(const Vector3D &point);

	/**
	 * This method traverses an OcTree.
	 */
	bool traverse(OcTreeIterator *iter);

protected:

	/**
	 * This methods is needed for propagating data stored in a subclass of an <code>OcTreeNode</code>
	 *
	 * @param destination
	 *			the destination <code>OcTreeTraversalNode</code>.
	 * @param source
	 *			the source <code>OcTreeTraversalNode</code>.
	 */
    virtual void copyTraversalNode(OcTreeTraversalNode *destination, OcTreeTraversalNode *source);

	/**
	 * This methods initializes an <code>OcTreeTraversalNode</code>
	 * @param newNode
	 *			is the new traversal node. 
	 * @param parentNode
	 *			is the current node.
	 * @param child
	 *			is the child number.
	 */
    virtual void initTraversalNode(OcTreeTraversalNode *newNode, OcTreeTraversalNode *parentNode, int child) { return; }

	/**
	 * This methods initializes an <code>OcTreeTraversalNode</code>
	 * @param travNode
	 *			reference to the traversal node.
	 * @param iter
	 *			reference to the iterator.
	 */
	bool traverseNode(OcTreeTraversalNode *travNode, OcTreeIterator *iter);

public:
    
	/// Root is at depth 0
	int 						maxActualDepth;	
    /// The center of an octree
	Vector3D					center;
	/// Half side length of the octree
	float						length;
	/// Reference to the root element
	OcTreeNode 					*root;
	/// Reference to the node factory
    OcTreeTraversalNodeFactory	*nodeFactory;
	/// Number of possible levels in the octree
	int							nLevels;		
	/// Level of root cell (nLevels - 1)
	int							rootLevel;		
	/// For converting positions to locational codes (maxVal = 2^rootLevel)
	float						maxVal;			
												
};


const int octreeChildCenterOffsetInt[][3] = {
	{-1, -1, -1},
	{ 1, -1, -1},
	{-1,  1, -1},
	{ 1,  1, -1},
	{-1, -1,  1},
	{ 1, -1,  1},
	{-1,  1,  1},
	{ 1,  1,  1}
};


const Vector3D octreeChildCenterOffset[] = {
    Vector3D(-1, -1, -1),
    Vector3D( 1, -1, -1),
    Vector3D(-1,  1, -1),
    Vector3D( 1,  1, -1),
    Vector3D(-1, -1,  1),
    Vector3D( 1, -1,  1),
    Vector3D(-1,  1,  1),
    Vector3D( 1,  1,  1)
};


/**
 * The OcTreeNode is a base class for a node of an <code>OcTree</code>.
 */
class OcTreeNode {

public:
	
	/**
	 * Constructor sets all locational codes to zero and all references to NULL.
	 */
    OcTreeNode() {
		xLocCode = 0;
		yLocCode = 0;
		zLocCode = 0;
		children = NULL;
		parent = NULL;
	}
	
	/**
	 * Destructor deletes all children recursively.
	 */
    ~OcTreeNode() {
        parent = NULL;
        if (children != NULL) {
            for(int i = 0; i < 8; i++) {
            	delete children[i];
			}
		}
    }
		
	/**
	 * Write coordinates to the vector point (according to the locational codes).
	 * Coordinates are element of [0,1] x [0,1] x [0,1].
	 * 
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 */
	virtual void getCoordinates(Vector3D &point, OcTree *ocTree);
	
	/**
	 * This method returns a pointer of the left neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the left neighbour or NULL.
	 */
	OcTreeNode *getNeighbourLeft(OcTree *ocTree);
	
	/**
	 * This method returns a pointer to the left neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 *
	 * <b>Note:</b> The neighbour has to be on the same level as the current node.
	 *              Otherwise a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the left neighbour with same level or NULL.
	 */
	OcTreeNode *getNeighbourLeftSameLevel(OcTree *ocTree);
	
	/**
	 * This method returns a pointer of the right neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the right neighbour or NULL.
	 */
	OcTreeNode *getNeighbourRight(OcTree *ocTree);

	/**
	 * This method returns a pointer to the right neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 *
	 * <b>Note:</b> The neighbour has to be on the same level as the current node.
	 *              Otherwise a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the right neighbour with same level or NULL.
	 */
	OcTreeNode *getNeighbourRightSameLevel(OcTree *ocTree);

	/**
	 * This method returns a pointer of the front neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the front neighbour or NULL.
	 */
	OcTreeNode *getNeighbourFront(OcTree *ocTree);

	/**
	 * This method returns a pointer to the front neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 *
	 * <b>Note:</b> The neighbour has to be on the same level as the current node.
	 *              Otherwise a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the front neighbour with same level or NULL.
	 */
	OcTreeNode *getNeighbourFrontSameLevel(OcTree *ocTree);

	/**
	 * This method returns a pointer of the rear neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the rear neighbour or NULL.
	 */	
	OcTreeNode *getNeighbourRear(OcTree *ocTree);

	/**
	 * This method returns a pointer to the rear neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 *
	 * <b>Note:</b> The neighbour has to be on the same level as the current node.
	 *              Otherwise a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the rear neighbour with same level or NULL.
	 */
	OcTreeNode *getNeighbourRearSameLevel(OcTree *ocTree);

	/**
	 * This method returns a pointer of the bottom neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the bottom neighbour or NULL.
	 */	
	OcTreeNode *getNeighbourBottom(OcTree *ocTree);

	/**
	 * This method returns a pointer to the bottom neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 *
	 * <b>Note:</b> The neighbour has to be on the same level as the current node.
	 *              Otherwise a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the bottom neighbour with same level or NULL.
	 */
	OcTreeNode *getNeighbourBottomSameLevel(OcTree *ocTree);

	/**
	 * This method returns a pointer of the top neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the top neighbour or NULL.
	 */	
	OcTreeNode *getNeighbourTop(OcTree *ocTree);

	/**
	 * This method returns a pointer to the top neighbour. If no neighbour is found,
	 * a NULL pointer is returned.
	 *
	 * <b>Note:</b> The neighbour has to be on the same level as the current node.
	 *              Otherwise a NULL pointer is returned.
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to the top neighbour with same level or NULL.
	 */
	OcTreeNode *getNeighbourTopSameLevel(OcTree *ocTree);

	/**
	 * This method returns all 26 neighbours of the current OcTree node to.
	 * Nodes are sorted lexicographically from left to right, front to rear
	 * and bottom to top.
	 *
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to an array of neighbours or NULL.
	 */
	virtual OcTreeNode **getNeighbourCells(OcTree *ocTree);

	/**
	 * This method returns all 26 neighbours of same level the current OcTree node to.
	 * Nodes are sorted lexicographically from left to right, front to rear
	 * and bottom to top.
	 *
	 * <b>Note:</b> If a neighbour does not have the same level as the current
	 *				node, a NULL pointer is stored instead.
	 *
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns a pointer to an array of neighbours or NULL.
	 */
	virtual OcTreeNode **getNeighbourCellsSameLevel(OcTree *ocTree);
	
	/**
	 * Testing methods for the neighbour cells methods.
	 * 
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns true if test succesful.
	 */
	virtual bool getNeighbourCellsTest(OcTree *ocTree);

	/**
	 * Testing methods for the neighbour cells methods with same level.
	 * 
	 * @param ocTree
	 *				specifies the <code>OcTree</code> the node belongs to.
	 * @return returns true if test succesful.
	 */
	virtual bool getNeighbourCellsSameLevelTest(OcTree *ocTree);

public:
    /// pointer to the parent
	OcTreeNode		*parent;
    /// pointer to all 8 children (lexicographically sorted)
	OcTreeNode 		**children;		
    /// X locational code for neighbour searching
	unsigned short  xLocCode;
    /// Y locational code for neighbour searching
	unsigned short  yLocCode;
	/// Z locational code for neighbour searching
	unsigned short  zLocCode;
	/// actual level
	unsigned short	level;
};


/**
 * Used for temporary storage of octree nodes during traversal or insertion.
 */
class OcTreeTraversalNode {
public:
	/// current center
	Vector3D	center;
    /// current lenght
	float		length;
    /// current depth
	int			depth;			
	/// reference to the actual node
    OcTreeNode	*node;
};


/**
 * A factory for traversal nodes.
 */
class OcTreeTraversalNodeFactory {

public:
    virtual OcTreeTraversalNode *newTraversalNode() = NULL;
    virtual void freeTraversalNode(OcTreeTraversalNode *node) = NULL;
    virtual int getAllocatedCount(void) = NULL;
};


/**
 * This class is used for traversing over an </code>OcTree</code>. The OcTree
 * traversal method calls the iterate method of the iterator for every node in the
 * OcTree.
 *
 * <b>Note:</b> If showInternalNodes is set to false, the OcTreeIterator will only
 *				process leaf nodes.
 */
class OcTreeIterator {
public:
    virtual bool iterate(OcTreeTraversalNode *np) { return false; }

public:
	/// toggle show internal nodes
    bool showInternalNodes;
};


/**
 * An iterator to gather statistical data.
 */
class OcTreeStatistics : public OcTreeIterator {

public:
    OcTreeStatistics() {
        ocTree = NULL;
        leafCount = NULL;
        internalNodeCount = NULL;
        showInternalNodes = true;
    }

    ~OcTreeStatistics() {
        if (leafCount != NULL)
            delete[] leafCount;
        if (internalNodeCount != NULL)
            delete[] internalNodeCount;
    }

    void collectStatistics(OcTree *op);
    virtual void reportStatistics();
    virtual void reportStatisticsAtDepth(int d);
    virtual void clear();
    virtual bool iterate(OcTreeTraversalNode *np);

public:
	/// reference to the <code>OcTree</code>
    OcTree 	*ocTree;
	/// number of total nodes
    int 	totalNodes;
	/// number of leafs per level
    int		*leafCount;
	/// number of internal nodes per level
    int		*internalNodeCount;
};


/**
 * A base factory for <code>OcTreeTraversalNode</code>s.
 */
template <class T> class OcTreeTraversalNodePool : public OcTreeTraversalNodeFactory {

public:
	/**
	 * Constructor
	 *
	 * @param chunkSize
	 *				defines the size of a chunk.
	 * @param numChunks
	 *				defines the number of chunks.
	 * @param isGrowable
	 *				determines whether the pool is growable.
	 */
	OcTreeTraversalNodePool(int chunkSize, int numChunks, bool isGrowable = true)
    	: nodePool(chunkSize, numChunks, isGrowable) {}

	/**
	 * Allocate a new node from the pool.
	 * @return
	 *				a new traversal node.
	 */
    virtual OcTreeTraversalNode *newTraversalNode() { return nodePool.alloc(); }

	/**
	 * Deallocates a node from the pool.
	 */
    virtual void freeTraversalNode(OcTreeTraversalNode *travNode) { nodePool.dealloc( (T *)travNode ); }

	/**
	 * @return the number of allocated elements in the pool.
	 */
	virtual int getAllocatedCount(void) { return nodePool.getAllocatedCount(); }

public:
	/// the node pool
    pool<T> nodePool;
};

#endif	// __OCTREE_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
