// Title:   VolGrid.cpp
// Created: Thu Sep 25 15:10:15 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "VolGrid.h"


void VolGrid::createVolGrid(Vector3D& bBoxMin, Vector3D& bBoxMax, float bBoxSize, int maxDepth, int narrowBandNofVoxels) {

	int i;

	this->nofVoxels = narrowBandNofVoxels;

	// initialize variables
	this->maxDepth = maxDepth;

	// store bounding box 
	this->boundingBoxMin = bBoxMin;
	this->boundingBoxMax = bBoxMax;
	this->center = this->boundingBoxMin + 0.5 * (this->boundingBoxMax - this->boundingBoxMin);

	// init slice
	this->slice = new float[int(pow(pow(2, this->maxDepth), 2))];
	this->boundary = new float[int(pow(pow(2, this->maxDepth), 2))];

  	
	// length is half side length of the cube (plus the narrow band width)
	this->length =  0.5 * (MAX((this->boundingBoxMax[0] - this->boundingBoxMin[0]),
		                   MAX((this->boundingBoxMax[1] - this->boundingBoxMin[1]),
                     		   (this->boundingBoxMax[2] - this->boundingBoxMin[2]))));

	// add narrow band
	this->length += sqrt(((pow(2, this->maxDepth - 1) * ((float) narrowBandNofVoxels / 2))));
	this->length *= bBoxSize;

	// calculate the narrow band width
    this->narrowBandWidth = calculateNarrowBandWidth(narrowBandNofVoxels);

	// precalculate length table for faster access
   	this->lengthTable[0] = 2 * this->length;
  	for (i = 1; i < 16; i++)
  		this->lengthTable[i] = 0.5 * this->lengthTable[i - 1];

  	// create the top level node
	// localization codes are all set to zeroes
  	VolGridNode *node = new VolGridNode();
  	node->parent = NULL;
  	node->children = NULL;
  	node->nodeType = LEAF;
  	node->level = this->rootLevel;
	node->xLocCode = 0x00000000;
	node->yLocCode = 0x00000000;
	node->zLocCode = 0x00000000;
	node->boundaryNode = false;
	
	for (i = 0; i < 8; i++) {
		node->distanceValues[i] = NULL;
		node->diffusionValues[i] = NULL;
	}
	this->root = node;
	
  	VolGridTraversalNode travNode;
  	travNode.center = this->center;
  	travNode.length = this->length;
  	travNode.depth = 0;
  	travNode.node = root;

  	nodeFactory = new OcTreeTraversalNodePool<VolGridTraversalNode>(32, 1);

	// create the data using the builder
  	VolGridBuilder *builder = new VolGridBuilder(this);

	// call the VolGrid builder
	builder->build();
	delete builder;

}


float VolGrid::calculateNarrowBandWidth(int narrowBandNofVoxels) {

	if (narrowBandNofVoxels > 0) {
		return (this->length * pow(0.5, maxDepth - 1) * narrowBandNofVoxels);
	}
	else {
		return (this->lengthTable[this->maxDepth - 1]);
	}
}


bool VolGrid::getSlice(int level, int sliceNumber, PLANE plane) {

	int dim;
	int x, y;
	Vector3D point;
    OcTreeTraversalNode *travNode;

	travNode = nodeFactory->newTraversalNode();
    travNode->center = this->center;
    travNode->length = this->length;
    travNode->depth = 0;
    travNode->node = root;
    this->initTraversalNode(travNode, travNode, -1);
	
	dim = this->getSliceDimension(level);

	if ((level > this->maxDepth) || (sliceNumber > (dim - 1)))
		return false;

	// set the constant value
	switch (plane) {
		case (XY_PLANE):
			point[2] = (float) (sliceNumber + 0.5) / dim;
			break;
		case (XZ_PLANE):
			point[1] = (float) (sliceNumber + 0.5) / dim;
			break;
		case (YZ_PLANE):
			point[0] = (float) (sliceNumber + 0.5) / dim;
			break;
		default:
			point[0] = point[1] = point[2] = 0.0;
			break;
	}

	for (y = 0; y < dim; y++) {
		for (x = 0; x < dim; x++) {
			// locate the corresponding grid point
			switch (plane) {
				case (XY_PLANE):
					point[0] = (float) (x + 0.5) / dim;
					point[1] = (float) (y + 0.5) / dim;
					break;
				case (XZ_PLANE):
					point[0] = (float) (x + 0.5) / dim;
					point[2] = (float) (y + 0.5) / dim;
					break;
				case (YZ_PLANE):
					point[1] = (float) (x + 0.5) / dim;
					point[2] = (float) (y + 0.5) / dim;
					break;
				default:
					point[0] = point[1] = point[2] = 0.0;
					break;
			}

			if (this->getCellForPoint(point, travNode)) {
				VolGridNode *volNode = (VolGridNode *) (travNode->node);				
				float value = this->getNarrowBandWidth();

					if (volNode->nodeType == LEAF) {
						
						value = 0;
						int counter = 0;

						for (int i = 0; i < 8; i++) {
							if ((volNode->distanceValues[i]) != NULL && ((*volNode->distanceValues[i]) != FLT_MAX)) {
								value = value + *(volNode->distanceValues[i]);
								counter++;
							}
						}
						
						if (counter == 8) {
							value /= counter;
						}
						else {
							value = this->getNarrowBandWidth();
						}

					}
								
				this->slice[x + y*dim] = value;
				
				if (volNode->boundaryNode) {
					this->boundary[x + y*dim] = 1.0f;
				}
//				else if (volNode->boundaryNode && (volNode->valid)) {
//					this->boundary[x + y*dim] = 2.0f;
//				}
				else {
					this->boundary[x + y*dim] = 0.0f;
				}
			}
		}
	}

	nodeFactory->freeTraversalNode(travNode);
	
	return true;
}


void VolGridBuilder::initNearestNeighbours() {

	Scene            *scene;
	Object           *object;
	SurfelCollection *surfels;
	SurfelInterface  *surfel;
	int				 i;

	// Get the active object and build up the kd-tree.
	scene = Scene::getInstance();
	object = scene->getActiveObject();

	if (object) {
		surfels = object->getSurfelCollection();
		int nofSurfels = surfels->getNofSurfels();
		this->nofSurfels = nofSurfels;

		this->surfelPositions = new Vector3D [nofSurfels];
		this->surfelNormals = new Vector3D [nofSurfels];

		for (i = 0, surfel = surfels->getFirstSurfel(); i < nofSurfels; i++, surfel = surfels->getNextSurfel()) {
			this->surfelPositions[i] = surfel->getPosition();
			this->surfelNormals[i] = surfel->getNormal();
		}

		this->search = new NeighbourHood (this->surfelPositions, nofSurfels);
	}
}


bool VolGridBuilder::getDistanceFromCenter(VolGridTraversalNode* travNode) {

	bool ret;
	Vector3D toPoint;

	VolGridNode *node = (VolGridNode *) travNode->node;

	this->search->setNofNeighbours(10);
	this->search->setSourcePoint(travNode->center);

    // caculate distance vector
	toPoint = travNode->center - this->search->getNeighbourPosition(0);

    ret = true;
	
	// check if inside const*cube + narrowBandWidth
	float length =  1.25 * travNode->length + this->grid->narrowBandWidth;
    if ((fabs(toPoint[0]) > length) || (fabs(toPoint[1]) > length) || (fabs(toPoint[2]) > length)) {
		ret = false;
	}

	return ret;

/*

  	// other method (slower but more accurate)

	bool ret;
	Vector3D toPoint;

	VolGridNode *node = (VolGridNode *) travNode->node;

	this->search->setNofNeighbours(10);
	this->search->setSourcePoint(travNode->center);

    // caculate distance vector
	toPoint = travNode->center - this->search->getNeighbourPosition(0);

    ret = true;
		
	float dist;
	int neighbourHoodSize = this->grid->distanceFieldNofNeighbours;

	DISTANCEFIELDCALC method = this->grid->distanceFieldCalcMethod;
	if (this->nofSurfels <= neighbourHoodSize) {
		method = NEXT_POINT;
	}

	this->search->setNofNeighbours(neighbourHoodSize);

	switch (method) {
		case TANGENTPLANES:
			dist = getDistanceTangentPlane(travNode, travNode->center, neighbourHoodSize);
			break;
		case WEIGHTED_LEASTSQUARES:
			dist = getDistanceWeightedLeastSquares(travNode, travNode->center, neighbourHoodSize);
			break;				
		case OPT_WEIGHTED_LEASTSQUARES:
			dist = getDistanceOptWeightedLeastSquares(travNode, travNode->center, neighbourHoodSize);
			break;
		case NEXT_POINT:
			dist = getDistanceNextPoint(travNode, travNode->center);
			break;
		default:
			dist = getDistanceTangentPlane(travNode, travNode->center, neighbourHoodSize);
			break;
	}


	if (dist > (sqrt(3)*travNode->length + this->grid->narrowBandWidth)) {
		ret = false;
	}
	
	return ret;
*/
}


float VolGridBuilder::getDistanceWeightedLeastSquares(VolGridTraversalNode *travNode, Vector3D queryPos, int neighbourHoodSize) {
	gsl_matrix *A,*Q;
	gsl_vector *x, *b, *S,*work;
	
	unsigned int i;

	this->search->setNofNeighbours(neighbourHoodSize);
	this->search->setSourcePoint(queryPos);
	
	// calculate weights
	Vector3D nearestPoint = this->search->getNeighbourPosition(0);
	Vector3D queryPosNormal = (queryPos - this->search->getNeighbourPosition(0));
	queryPosNormal.normalize();
	Vector3D farthestPoint = this->search->getNeighbourPosition(neighbourHoodSize - 1);
	float sigma = (Vector3D(farthestPoint - nearestPoint).getLength()) / 3.0;
	float sigmaSquared = 2* (sigma * sigma);

	int size = 4;
	// for singular value decomposition.
	A = gsl_matrix_alloc(size, size);
	Q = gsl_matrix_alloc(size, size);
	S = gsl_vector_alloc(size);
	work = gsl_vector_alloc(size);
	x = gsl_vector_alloc(size);
	b = gsl_vector_alloc(size);

	gsl_matrix_set_zero(A);
	gsl_vector_set_zero(b);

	this->search->setNofNeighbours(neighbourHoodSize);
	this->search->setSourcePoint(nearestPoint);


	// a plane is fitted through points in the neighbourhood
	// all points have a gaussian weight exp(-distanceSquared/SigmaSquared)
	float normalWeight = 0.0f;
	Vector3D normalDirection = Vector3D(0.0f, 0.0f, 0.0f);

	for(i = 0; i < neighbourHoodSize; i++) {

		Vector3D pos = this->search->getNeighbourPosition(i);
		Vector3D normal = this->surfelNormals[this->search->getNeighbourPositionIndex(i)];

		// not defined ( != 1.0 does not work)
		if (normal.getLength() > 1.1 || normal.getLength() < 0.9) {
			continue;
		}

		float x = pos[0];
		float y = pos[1];
		float z = pos[2];

		float vx = x - nearestPoint[0];
		float vy = y - nearestPoint[1];
		float vz = z - nearestPoint[2];

		float distanceSquared = sqrt(vx*vx + vy*vy + vz*vz);
		float w = exp(-distanceSquared / sigmaSquared);

		gsl_matrix_set(A, 0, 0, gsl_matrix_get(A, 0, 0) + w);
		gsl_matrix_set(A, 0, 1, gsl_matrix_get(A, 0, 1) + w*x);
		gsl_matrix_set(A, 0, 2, gsl_matrix_get(A, 0, 2) + w*y);
		gsl_matrix_set(A, 0, 3, gsl_matrix_get(A, 0, 3) + w*z);

		gsl_matrix_set(A, 1, 0, gsl_matrix_get(A, 1, 0) + w*x);
		gsl_matrix_set(A, 1, 1, gsl_matrix_get(A, 1, 1) + w*x*x);
		gsl_matrix_set(A, 1, 2, gsl_matrix_get(A, 1, 2) + w*y*x);
		gsl_matrix_set(A, 1, 3, gsl_matrix_get(A, 1, 3) + w*z*x);

		gsl_matrix_set(A, 2, 0, gsl_matrix_get(A, 2, 0) + w*y);
		gsl_matrix_set(A, 2, 1, gsl_matrix_get(A, 2, 1) + w*x*y);
		gsl_matrix_set(A, 2, 2, gsl_matrix_get(A, 2, 2) + w*y*y);
		gsl_matrix_set(A, 2, 3, gsl_matrix_get(A, 2, 3) + w*z*y);

		gsl_matrix_set(A, 3, 0, gsl_matrix_get(A, 3, 0) + w*z);
		gsl_matrix_set(A, 3, 1, gsl_matrix_get(A, 3, 1) + w*x*z);
		gsl_matrix_set(A, 3, 2, gsl_matrix_get(A, 3, 2) + w*y*z);
		gsl_matrix_set(A, 3, 3, gsl_matrix_get(A, 3, 3) + w*z*z);
		
		gsl_vector_set(b, 0, gsl_vector_get(b,0) + w);
		gsl_vector_set(b, 1, gsl_vector_get(b,1) + w);
		gsl_vector_set(b, 2, gsl_vector_get(b,2) + w);
		gsl_vector_set(b, 3, gsl_vector_get(b,3) + w);


		distanceSquared = Vector3D(nearestPoint - this->search->getNeighbourPosition(i)).getSquaredLength();
		normalWeight = exp(-distanceSquared / (sigmaSquared));
		normalDirection += normalWeight * normal;
	
	}
	gsl_linalg_SV_decomp(A,Q,S,work);
//	gsl_linalg_SV_decomp_jacobi (A, Q, S);
	gsl_linalg_SV_solve(A,Q,S,b,x);

	float _d = gsl_vector_get(x, 0);
	float _a = gsl_vector_get(x, 1);
	float _b = gsl_vector_get(x, 2);
	float _c = gsl_vector_get(x, 3);

	double len = sqrt(_a*_a + _b*_b + _c*_c);
	if(len != 0){
		_a = (float)(_a/len);
		_b = (float)(_b/len);
		_c = (float)(_c/len);
		_d = (float)(_d/len);
	}

	gsl_vector_free(S);
	gsl_vector_free(work);
	gsl_vector_free(b);
	gsl_vector_free(x);
	gsl_matrix_free(A);
	gsl_matrix_free(Q);

	Vector3D normal = Vector3D(_a, _b, _c);
	Vector3D anchor;
	float value;

	normalDirection.normalize();
	float sign = 1.0f;
	
	if (Vector3D::dotProduct(normalDirection, queryPosNormal) < 0.0f) {
		sign = -1.0f;
	}

	if (_a != 0.0f) {
		anchor = Vector3D(-_d/_a, 0, 0);
	}
	else if (_b != 0.0f) {
		anchor = Vector3D(0, -_d/_b, 0);
	}
	else if (_c != 0.0f) {
		anchor = Vector3D(0, 0, -_d/_c);
	}
	else {
		return sign * Vector3D::distance(nearestPoint, queryPos);
	}

	value = sign * Vector3D(Vector3D::dotProduct(normal, queryPos - anchor)*normal).getLength();

	return value;
}


float VolGridBuilder::getDistanceOptWeightedLeastSquares(VolGridTraversalNode *travNode, Vector3D queryPos, int neighbourHoodSize) {

	gsl_matrix *A,*Q;
	gsl_vector *x, *b, *S,*work;
	unsigned int i;

	this->search->setSourcePoint(queryPos);
	
	// Calculate sigma
	Vector3D nearestPoint = this->search->getNeighbourPosition(0);
	Vector3D queryPosNormal = (queryPos - this->search->getNeighbourPosition(0));
	queryPosNormal.normalize();
	Vector3D farthestPoint = this->search->getNeighbourPosition(neighbourHoodSize - 1);
	float sigma = (Vector3D(farthestPoint - nearestPoint).getLength()) / 3.0;
	float sigmaSquared = 2*(sigma * sigma);
	
	// Allocate matrices for singular value decomposition.
	A = gsl_matrix_alloc(4, 4);
	Q = gsl_matrix_alloc(4, 4);
	S = gsl_vector_alloc(4);
	work = gsl_vector_alloc(4);
	x = gsl_vector_alloc(4);
	b = gsl_vector_alloc(4);

	gsl_matrix_set_zero(A);
	gsl_vector_set_zero(b);

	// Set source point
	this->search->setSourcePoint(nearestPoint);

	// Offsetting is defined by the octree cell length
	float off = 0.01f * travNode->length;

	// A plane is fitted through points in the neighbourhood
	// All points have a gaussian weight exp(-distanceSquared/(2*sigma*sigma))
	float normalWeight = 0.0f;
	Vector3D normalDirection = Vector3D(0.0f, 0.0f, 0.0f);

	for(i = 0; i < neighbourHoodSize; i++) {

		Vector3D pos = this->search->getNeighbourPosition(i);
		Vector3D normal = this->surfelNormals[this->search->getNeighbourPositionIndex(i)];

		// not defined ( != 1.0 does not work)
		if (normal.getLength() > 1.1 || normal.getLength() < 0.9) {
			continue;
		}

		float x = pos[0] + off*normal[0];
		float y = pos[1] + off*normal[1];
		float z = pos[2] + off*normal[2];

		float vx = x - nearestPoint[0];
		float vy = y - nearestPoint[1];
		float vz = z - nearestPoint[2];

		float distanceSquared = sqrt(vx*vx + vy*vy + vz*vz);
		float w = exp(-distanceSquared / sigmaSquared);

		gsl_matrix_set(A, 0, 0, gsl_matrix_get(A, 0, 0) + w);
		gsl_matrix_set(A, 0, 1, gsl_matrix_get(A, 0, 1) + w*x);
		gsl_matrix_set(A, 0, 2, gsl_matrix_get(A, 0, 2) + w*y);
		gsl_matrix_set(A, 0, 3, gsl_matrix_get(A, 0, 3) + w*z);

		gsl_matrix_set(A, 1, 0, gsl_matrix_get(A, 1, 0) + w*x);
		gsl_matrix_set(A, 1, 1, gsl_matrix_get(A, 1, 1) + w*x*x);
		gsl_matrix_set(A, 1, 2, gsl_matrix_get(A, 1, 2) + w*y*x);
		gsl_matrix_set(A, 1, 3, gsl_matrix_get(A, 1, 3) + w*z*x);

		gsl_matrix_set(A, 2, 0, gsl_matrix_get(A, 2, 0) + w*y);
		gsl_matrix_set(A, 2, 1, gsl_matrix_get(A, 2, 1) + w*x*y);
		gsl_matrix_set(A, 2, 2, gsl_matrix_get(A, 2, 2) + w*y*y);
		gsl_matrix_set(A, 2, 3, gsl_matrix_get(A, 2, 3) + w*z*y);

		gsl_matrix_set(A, 3, 0, gsl_matrix_get(A, 3, 0) + w*z);
		gsl_matrix_set(A, 3, 1, gsl_matrix_get(A, 3, 1) + w*x*z);
		gsl_matrix_set(A, 3, 2, gsl_matrix_get(A, 3, 2) + w*y*z);
		gsl_matrix_set(A, 3, 3, gsl_matrix_get(A, 3, 3) + w*z*z);
		
		gsl_vector_set(b, 0, gsl_vector_get(b,0) + w*off);
		gsl_vector_set(b, 1, gsl_vector_get(b,1) + w*x*off);
		gsl_vector_set(b, 2, gsl_vector_get(b,2) + w*y*off);
		gsl_vector_set(b, 3, gsl_vector_get(b,3) + w*z*off);

		x = pos[0] - off*normal[0];
		y = pos[1] - off*normal[1];
		z = pos[2] - off*normal[2];

		vx = x - nearestPoint[0];
		vy = y - nearestPoint[1];
		vz = z - nearestPoint[2];
		
		distanceSquared = sqrt(vx*vx + vy*vy + vz*vz);
		w = exp(-distanceSquared / sigmaSquared);

		gsl_matrix_set(A, 0, 0, gsl_matrix_get(A, 0, 0) + w);
		gsl_matrix_set(A, 0, 1, gsl_matrix_get(A, 0, 1) + w*x);
		gsl_matrix_set(A, 0, 2, gsl_matrix_get(A, 0, 2) + w*y);
		gsl_matrix_set(A, 0, 3, gsl_matrix_get(A, 0, 3) + w*z);

		gsl_matrix_set(A, 1, 0, gsl_matrix_get(A, 1, 0) + w*x);
		gsl_matrix_set(A, 1, 1, gsl_matrix_get(A, 1, 1) + w*x*x);
		gsl_matrix_set(A, 1, 2, gsl_matrix_get(A, 1, 2) + w*y*x);
		gsl_matrix_set(A, 1, 3, gsl_matrix_get(A, 1, 3) + w*z*x);

		gsl_matrix_set(A, 2, 0, gsl_matrix_get(A, 2, 0) + w*y);
		gsl_matrix_set(A, 2, 1, gsl_matrix_get(A, 2, 1) + w*x*y);
		gsl_matrix_set(A, 2, 2, gsl_matrix_get(A, 2, 2) + w*y*y);
		gsl_matrix_set(A, 2, 3, gsl_matrix_get(A, 2, 3) + w*z*y);

		gsl_matrix_set(A, 3, 0, gsl_matrix_get(A, 3, 0) + w*z);
		gsl_matrix_set(A, 3, 1, gsl_matrix_get(A, 3, 1) + w*x*z);
		gsl_matrix_set(A, 3, 2, gsl_matrix_get(A, 3, 2) + w*y*z);
		gsl_matrix_set(A, 3, 3, gsl_matrix_get(A, 3, 3) + w*z*z);
		
		gsl_vector_set(b, 0, gsl_vector_get(b,0) - w*off);
		gsl_vector_set(b, 1, gsl_vector_get(b,1) - w*x*off);
		gsl_vector_set(b, 2, gsl_vector_get(b,2) - w*y*off);
		gsl_vector_set(b, 3, gsl_vector_get(b,3) - w*z*off);

		distanceSquared = Vector3D(nearestPoint - this->search->getNeighbourPosition(i)).getSquaredLength();
		normalWeight = exp(-distanceSquared / sigmaSquared);
		normalDirection += normalWeight * normal;
	
	}

	// SVD Decomposition
	gsl_linalg_SV_decomp(A,Q,S,work);
	//	gsl_linalg_SV_decomp_jacobi (A, Q, S);			alternative
	gsl_linalg_SV_solve(A,Q,S,b,x);

	float ax = gsl_vector_get(x, 1);
	float by = gsl_vector_get(x, 2);
	float cz = gsl_vector_get(x, 3);
	float  d = gsl_vector_get(x, 0);

	double len = sqrt(ax*ax + by*by + cz*cz);
	if(len != 0){
		ax = (float)(ax/len);
		by = (float)(by/len);
		cz = (float)(cz/len);
		d =  (float)( d/len);
	}

	gsl_vector_free(S);
	gsl_vector_free(work);
	gsl_vector_free(b);
	gsl_vector_free(x);
	gsl_matrix_free(A);
	gsl_matrix_free(Q);

	Vector3D normal = Vector3D(ax, by, cz);
	Vector3D anchor;
	float value;

	normalDirection.normalize();
	
	if (ax != 0.0f) {
		anchor = Vector3D(-d/ax, 0, 0);
	}
	else if (by != 0.0f) {
		anchor = Vector3D(0, -d/by, 0);
	}
	else if (cz != 0.0f) {
		anchor = Vector3D(0, 0, -d/cz);
	}

	if (Vector3D::dotProduct(normal, queryPosNormal) < 0.0f) {
		normal *= -1;
	}

	value = Vector3D::dotProduct(normal, queryPos - anchor);

	if ((ax == 0.0) && (by == 0.0) && (cz == 0.0)) {
		value = Vector3D::distance(nearestPoint, queryPos);
	}

	int counter = 0;
	for (int k = 0; k < neighbourHoodSize; k++) {
		if (Vector3D::dotProduct (normal, this->surfelNormals[this->search->getNeighbourPositionIndex (k)]) < 0.0f) {
			counter++;
		}
	}

	if (counter > (neighbourHoodSize / 2)) { value = -value; }

	return value;
}


float VolGridBuilder::getDistanceTangentPlane(VolGridTraversalNode* travNode, Vector3D queryPos, int neighbourHoodSize) {

	Vector3D pos;
	this->search->setNofNeighbours(neighbourHoodSize);

	// set the new source point for the neighbour hood
	this->search->setSourcePoint(queryPos);
	pos = this->search->getNeighbourPosition(0);
	this->search->setSourcePoint(pos);

	// caculate distance vector
	Vector3D toPoint = queryPos - pos;

	// compute normal (without orientation)
	Vector3D normal = this->search->computeNormal();

	// compute distance to estimated tangent plane
	float dist = Vector3D::dotProduct(toPoint, normal);

	int counter = 0;
	for (int k = 0; k < neighbourHoodSize; k++) {
		if (Vector3D::dotProduct (normal, this->surfelNormals[this->search->getNeighbourPositionIndex (k)]) < 0.0f) {
			counter++;
		}
	}

	if (counter > (neighbourHoodSize / 2)) { dist = -dist; }

	return dist;
}


float VolGridBuilder::getDistanceNextPoint(VolGridTraversalNode *travNode, Vector3D queryPos) {
	
	Vector3D pos;
	this->search->setNofNeighbours(1);

	// set the new source point for the neighbour hood
	this->search->setSourcePoint(queryPos);

	pos = this->search->getNeighbourPosition(0);
	this->search->setSourcePoint(pos);

	// caculate distance vector
	Vector3D toPoint = queryPos - this->search->getNeighbourPosition(0);

	// compute distance to estimated tangent plane
	float dist = toPoint.getLength();

	if (Vector3D::dotProduct (toPoint, this->surfelNormals[this->search->getNeighbourPositionIndex (0)]) < 0.0f) {
		dist *= -1;
	}

	return dist;
}


void VolGridBuilder::getDistanceValues(VolGridTraversalNode* travNode) {

	Vector3D toPoint, vertex, normal;
	VolGridNode *node = (VolGridNode *) travNode->node;
	VolGridNode **neighbours;

	int i, j;

	neighbours = (VolGridNode **)travNode->node->getNeighbourCellsSameLevel((OcTree *)this->grid);

	// copy values from neighbour nodes as references
	// so no duplicate values exist
	for (i = 0; i < 8; i++) {

		int	nbNode[27];
		int distValue[27];
		int nbCopyNode[7];
		int nbCopyNodePos[7];
		neighbourDistancePositions *nbrDisPos = &(neighbourDistanceTable[i]);
		
		memcpy(nbNode, nbrDisPos->neighbourNode, 27 * sizeof(int));
		memcpy(distValue, nbrDisPos->distanceValueNumber, 27 * sizeof(int));	
		memcpy(nbCopyNode, nbrDisPos->nbCopyNode, 7 * sizeof(int));
		memcpy(nbCopyNodePos, nbrDisPos->nbCopyNodePos, 7 * sizeof(int));
		
		for (int p = 0; p < 7; p++) {
			
			VolGridNode *copyNode = neighbours[nbCopyNode[p]];

			if ((copyNode) && (((VolGridNode *)travNode->node)->distanceValues[i] == NULL)) {
				((VolGridNode *)travNode->node)->distanceValues[i] = copyNode->distanceValues[nbCopyNodePos[p]];
				((VolGridNode *)travNode->node)->diffusionValues[i] = copyNode->diffusionValues[nbCopyNodePos[p]];
			}
			else if (((VolGridNode *)travNode->node)->distanceValues[i] != NULL) {
				break;
			}
		}
		
	}

	delete []neighbours;
	neighbours = NULL;

	for (i = 0; i < 8; i++) {
		if (((VolGridNode *)travNode->node)->distanceValues[i] == NULL) {

			for (j = 0; j < 3; j++) {
				vertex[j] = travNode->center[j] + (travNode->length * octreeChildCenterOffsetInt[i][j]);
			}
			
			float *dist = new float(0.0f);
			int neighbourHoodSize = this->grid->distanceFieldNofNeighbours;

			DISTANCEFIELDCALC method = this->grid->distanceFieldCalcMethod;
			if (this->nofSurfels <= neighbourHoodSize) {
				method = NEXT_POINT;
			}

			this->search->setNofNeighbours(neighbourHoodSize);

			switch (method) {
				case TANGENTPLANES:
					*dist = getDistanceTangentPlane(travNode, vertex, neighbourHoodSize);
					break;
				case WEIGHTED_LEASTSQUARES:
					*dist = getDistanceWeightedLeastSquares(travNode, vertex, neighbourHoodSize);
					break;				
				case OPT_WEIGHTED_LEASTSQUARES:
					*dist = getDistanceOptWeightedLeastSquares(travNode, vertex, neighbourHoodSize);
					break;
				case NEXT_POINT:
					*dist = getDistanceNextPoint(travNode, vertex);
					break;
				default:
					*dist = getDistanceTangentPlane(travNode, vertex, neighbourHoodSize);
					break;
			}

			((VolGridNode *)travNode->node)->distanceValues[i] = dist;

			float *diff = new float(*dist);
			((VolGridNode *)travNode->node)->diffusionValues[i] = diff;
		}
	}
}


bool VolGridBuilder::shouldSplit(VolGridTraversalNode *travNode) {

    bool retVal;
	VolGridNode *node = (VolGridNode *) travNode->node;

	retVal = getDistanceFromCenter(travNode);

	if (travNode->depth == this->grid->maxDepth) {
		node->nodeType = LEAF;
		node->valid = true;
		return false;
	}

	if (!retVal) {
		node->nodeType = EXTERNAL;
		node->valid = true;
		return false;
	}

	// otherwise the node should be splitted
	node->nodeType = INTERNAL;
	node->valid = true;
    return true;
}


bool VolGridBuilder::iterate(OcTreeTraversalNode *ocTravNode) {

	int i, j;
	int nextLevel;

    VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

    // update the maxActualDepth
    if (travNode->depth > grid->maxActualDepth)
    	grid->maxActualDepth = travNode->depth;

	// calculate distance values for this node
    this->getDistanceValues(travNode);

	// check to see if the node needs to be split
    if (this->shouldSplit(travNode)) {

		// create new nodes
		node->children = (OcTreeNode **) new VolGridNode *[8];

		for (i = 0; i < 8; i++) {
			// init the childnode
			node->children[i] = new VolGridNode();
            node->children[i]->parent = node;

			for (j = 0; j < 8; j++) {
				// set all distance and diffusion values to NULL
				((VolGridNode *)node->children[i])->distanceValues[j] = NULL;
				((VolGridNode *)node->children[i])->diffusionValues[j] = NULL;
			}

			// copy already calculated values from parent
			if (node->parent) {
				((VolGridNode *)node->children[i])->distanceValues[i] = node->distanceValues[i];
				((VolGridNode *)node->children[i])->diffusionValues[i] = node->diffusionValues[i];
			}

			// set locational codes for neighbour searching
			nextLevel = node->level - 1;
			node->children[i]->level = nextLevel;
			node->children[i]->children = NULL;

			if (octreeChildCenterOffsetInt[i][0] == 1) {
				node->children[i]->xLocCode = (node->xLocCode) | (0x00000001 << nextLevel);
			}
			else {
				node->children[i]->xLocCode = node->xLocCode;
			}
			if (octreeChildCenterOffsetInt[i][1] == 1) {
				node->children[i]->yLocCode = (node->yLocCode) | (0x00000001 << nextLevel);
			}
			else {
				node->children[i]->yLocCode = node->yLocCode;
			}
			if (octreeChildCenterOffsetInt[i][2] == 1) {
				node->children[i]->zLocCode = (node->zLocCode) | (0x00000001 << nextLevel);
			}
			else {
				node->children[i]->zLocCode = node->zLocCode;
			}

			((VolGridNode *)node->children[i])->boundaryNode = false;
			((VolGridNode *) node->children[i])->valid = true;
            
			// mark the children nodes as leaf
			((VolGridNode *) node->children[i])->nodeType = LEAF;
        }
    }
	
    return true;
}


void VolGridBuilder::build() {

	// init the builder and then traverse the octree
	this->initNearestNeighbours();
	this->grid->traverse(this);
}


VolGridNode *VolGridNarrowBandIterator::createVolGridNode(VolGridNode *commonNode, unsigned int nextLevel,
														  unsigned int xLocCode, unsigned int yLocCode, unsigned int zLocCode) {

	int l;
	VolGridNode *node;
	VolGridNode **neighbours;
		
	node = commonNode;

	while ((nextLevel) >= (this->grid->rootLevel - this->grid->maxDepth)) {

		unsigned int childBranchBit = 1 << (nextLevel);
        unsigned int childIndex = ((((xLocCode) & childBranchBit) >> (nextLevel))
		+ (((yLocCode) & childBranchBit) >> ((nextLevel - 1)))
        + (((zLocCode) & childBranchBit) >> ((nextLevel - 2))));

		// create new nodes if neccessary
		if (!(node)->children) {

			node->children = (OcTreeNode **) new VolGridNode *[8];
			int i;
			for (i = 0; i < 8; i++) {
				// init the childnode
				node->children[i] = new VolGridNode();
				node->children[i]->parent = node;

				for (int j = 0; j < 8; j++) {
					// set all distance and diffusion values to NULL
					((VolGridNode *)node->children[i])->distanceValues[j] = NULL;
					((VolGridNode *)node->children[i])->diffusionValues[j] = NULL;
				}

				// copy already calculated values from parent
				if (node->parent) {
					((VolGridNode *)node->children[i])->distanceValues[i] = ((VolGridNode *)node)->distanceValues[i];
					((VolGridNode *)node->children[i])->diffusionValues[i] = ((VolGridNode *)node)->diffusionValues[i];
				}


				// set locational codes for neighbour searching
				node->children[i]->level = nextLevel;
				node->children[i]->children = NULL;

				if (octreeChildCenterOffsetInt[i][0] == 1) {
					node->children[i]->xLocCode = (node->xLocCode) | (0x00000001 << nextLevel);
				}
				else {
					node->children[i]->xLocCode = node->xLocCode;
				}
				if (octreeChildCenterOffsetInt[i][1] == 1) {
					node->children[i]->yLocCode = (node->yLocCode) | (0x00000001 << nextLevel);
				}
				else {
					node->children[i]->yLocCode = node->yLocCode;
				}
				if (octreeChildCenterOffsetInt[i][2] == 1) {
					node->children[i]->zLocCode = (node->zLocCode) | (0x00000001 << nextLevel);
				}
				else {
					node->children[i]->zLocCode = node->zLocCode;
				}

				// mark the children nodes as internal or as leaf
				((VolGridNode *) node->children[i])->boundaryNode = false;

				if (nextLevel == (this->grid->rootLevel - this->grid->maxDepth)) {
					((VolGridNode *) node->children[i])->nodeType = LEAF;
					((VolGridNode *) node->children[i])->valid = false;
				}
				else {
					((VolGridNode *) node->children[i])->nodeType = INTERNAL;
					((VolGridNode *) node->children[i])->valid = false;
				}

			}

			if (nextLevel == (this->grid->rootLevel - this->grid->maxDepth)) {
				((VolGridNode *) node->children[childIndex])->boundaryNode = true;
			}
			else {
				((VolGridNode *) node->children[childIndex])->nodeType = INTERNAL;
			}

			int counter = 0;
			for (i = 0; i < 8; i++) {
				// copy other values 
				
				neighbours = (VolGridNode **)node->children[i]->getNeighbourCellsSameLevel((OcTree *)this->grid);

				// copy values from neighbour nodes as references
				// so no duplicate values exist
				for (l = 0; l < 8; l++) {

					int	nbNode[27];
					int distValue[27];
					int nbCopyNode[7];
					int nbCopyNodePos[7];
					neighbourDistancePositions* nbrDisPos = &(neighbourDistanceTable[l]);
					
					memcpy(nbNode, nbrDisPos->neighbourNode, 27 * sizeof(int));
					memcpy(distValue, nbrDisPos->distanceValueNumber, 27 * sizeof(int));	
					memcpy(nbCopyNode, nbrDisPos->nbCopyNode, 7 * sizeof(int));
					memcpy(nbCopyNodePos, nbrDisPos->nbCopyNodePos, 7 * sizeof(int));

					for (int p = 0; p < 7; p++) {
						
						VolGridNode *copyNode = neighbours[nbCopyNode[p]];

						if ((copyNode) && (((VolGridNode *)node->children[i])->distanceValues[l] == NULL)) {
							((VolGridNode *)node->children[i])->distanceValues[l] = copyNode->distanceValues[nbCopyNodePos[p]];
							((VolGridNode *)node->children[i])->diffusionValues[l] = copyNode->diffusionValues[nbCopyNodePos[p]];
						}
						else if (((VolGridNode *)node->children[i])->distanceValues[l] != NULL) {
							break;
						}
					}
					// else create new values with FLT_MAX
					if (((VolGridNode *)node->children[i])->distanceValues[l] == NULL) {
						float *dist = new float(FLT_MAX);
						float *diff = new float(FLT_MAX);
						((VolGridNode *)node->children[i])->valid = false;
						((VolGridNode *)node->children[i])->distanceValues[l] = dist;
						((VolGridNode *)node->children[i])->diffusionValues[l] = diff;
					}
					else { counter++; }
				}

				delete []neighbours;
				neighbours = NULL;
				
				if (counter == 8) {
					((VolGridNode *)node->children[i])->valid = true;
				}
			}		
		
		}
		else if (nextLevel == (this->grid->rootLevel - this->grid->maxDepth)) {
			((VolGridNode *) node->children[childIndex])->boundaryNode = true;
		}
		else {
			((VolGridNode *) node->children[childIndex])->nodeType = INTERNAL;
		}


		--nextLevel;
		node = (VolGridNode *)node->children[childIndex];
	}

	return node;
}


void VolGridNarrowBandIterator::growVolGrid(int value) {

	int xDim, yDim, zDim;
	unsigned int x0LocCode, y0LocCode, z0LocCode;
	unsigned int x1LocCode, y1LocCode, z1LocCode;
	VolGridNode *commonNode, *newNode;
	int j, k, l;

	this->largestHole = value;
	
	value = (value + 1) / 2;
	std::set<class VolGridNode*>::iterator i;
	for (i = this->boundaryNodeList.begin(); i != this->boundaryNodeList.end(); i++) {

		unsigned int binaryCellSize = 1 << ((*i)->level);

		xDim = yDim = zDim = 0;

		// copy locational codes from actual node
		x0LocCode = x1LocCode = (*i)->xLocCode;
		y0LocCode = y1LocCode = (*i)->yLocCode;
		z0LocCode = z1LocCode = (*i)->zLocCode;
		
		for (int n = 0; n < value; n++) {
			
			if (!((x0LocCode - binaryCellSize) >= (1 << this->grid->rootLevel))) {
				x0LocCode -= binaryCellSize;
				xDim++;
			}
			if (!((y0LocCode - binaryCellSize) >= (1 << this->grid->rootLevel))) {
				y0LocCode -= binaryCellSize;
				yDim++;
			}
			if (!((z0LocCode - binaryCellSize) >= (1 << this->grid->rootLevel))) {
				z0LocCode -= binaryCellSize;
				zDim++;
			}
			if (!((x1LocCode + binaryCellSize) >= (1 << this->grid->rootLevel))) {
				x1LocCode += binaryCellSize;
				xDim++;
			}
			if (!((y1LocCode + binaryCellSize) >= (1 << this->grid->rootLevel))) {
				y1LocCode += binaryCellSize;
				yDim++;
			}
			if (!((z1LocCode + binaryCellSize) >= (1 << this->grid->rootLevel))) {
				z1LocCode += binaryCellSize;
				zDim++;
			}
		}

		// find out common node for all cells

		unsigned int storedXLocCode = x0LocCode;
		unsigned int storedYLocCode = y0LocCode;

		for (l = 0; l <= zDim; l++) {
			for (k = 0; k <= yDim; k++) {
				for (j = 0; j <= xDim; j++) {
/*
					// determine common ancestor
					unsigned int levelX, levelY, levelZ, nextLevel;
					OcTreeNode *commonX, *commonY, *commonZ;
					levelX = levelY = levelZ = (*i)->level;
					unsigned int diffX = (*i)->xLocCode ^ x0LocCode;
					unsigned int diffY = (*i)->yLocCode ^ y0LocCode;
					unsigned int diffZ = (*i)->zLocCode ^ z0LocCode;
					commonX = commonY = commonZ = *i;

					OT_GET_COMMON_ANCESTOR(commonX, levelX, diffX);
					OT_GET_COMMON_ANCESTOR(commonY, levelY, diffY);
					OT_GET_COMMON_ANCESTOR(commonZ, levelZ, diffZ);
				
					if ((levelX > levelY) && (levelX > levelZ)) {
						nextLevel = levelX - 1;
						commonNode = (VolGridNode *)commonX;
					}
					else if (levelY > levelZ) {
						nextLevel = levelY - 1;
						commonNode = (VolGridNode *)commonY;
					}
					else {
						nextLevel = levelZ - 1;
						commonNode = (VolGridNode *)commonZ;
					}
*/
					commonNode = (VolGridNode *)this->grid->root;
					unsigned int nextLevel = this->grid->rootLevel - 1;
					
					// create VolGrid node
					newNode = this->createVolGridNode(commonNode, nextLevel, x0LocCode, y0LocCode, z0LocCode);
					
					// insert node into diffusion and into narrow band list
					diffusionNodeList.insert(newNode);
					nodeList.insert(newNode);

					x0LocCode += binaryCellSize;
				}		
				x0LocCode = storedXLocCode;
				y0LocCode += binaryCellSize;
			}			
			y0LocCode = storedYLocCode;
			z0LocCode += binaryCellSize;
		}
	}

	for (i = this->boundaryNodeList.begin(); i != this->boundaryNodeList.end(); i++) {
		(*i)->valid = true;
	}
}


void VolGridNarrowBandIterator::findBoundaryNodes(int oppositeSign, int invalidNodes) {

	int m, n;
	int invalidNodesCount, oppositeSignCount;
	int boundaryNodesCount = 0;

	int distanceValueProjection[8];
	float neigbhourHoodGrid[64];
	VolGridNode *node;
	Vector3D point;
	VolGridNode **neighbours;
	VolGridTraversalNode *travNode = (VolGridTraversalNode *)this->grid->nodeFactory->newTraversalNode();
	
	for (std::set<class VolGridNode*>::iterator i = this->nodeList.begin(); i != this->nodeList.end(); i++) {

		neighbours = (VolGridNode **)(*i)->getNeighbourCellsSameLevel(this->grid);

		for (n = 0; n < 64; n++) {
			neigbhourHoodGrid[n] = FLT_MAX;
		}

		for (n = 0; n < 27; n++) {

			neighbourHoodInfo* nbrTable = &(neighbourHoodTable[n]);
			memcpy(distanceValueProjection, nbrTable->distanceValueProjection, 8 * sizeof(int));
				
			if (n < 13) {
				node = neighbours[n];
			}
			else if (n > 13) {
				node = neighbours[n - 1];
			}
			else {
				node = (*i);
			}

			if (node) {
				for (m = 0; m < 8; m++) {
					if (neigbhourHoodGrid[distanceValueProjection[m]] == FLT_MAX) {
						neigbhourHoodGrid[distanceValueProjection[m]] = *node->distanceValues[m];
					}
				}				
			}
		}

		delete []neighbours;
		neighbours = NULL;

		// fill nodes in the neighbourhood grid which are not filled in with direct neighbour information
		for (n = 0; n < 64; n++) {
			node = (*i);
			if (neigbhourHoodGrid[n] == FLT_MAX) {
				
				node->getCoordinates(point,this->grid);

				float length = this->grid->lengthTable[this->grid->maxDepth];
				point[0] = neighbourOffsets[n][0] * length;
				point[1] = neighbourOffsets[n][1] * length;
				point[2] = neighbourOffsets[n][2] * length;
				
				// look-up all eight possible neighbours
				// if there is a neighbour fill in the value and go to the next cell in the neigbourhood grid
				for (m = 0; m < 8; m++) {
					if (this->grid->getCellForPoint(point + octreeChildCenterOffset[m] * length, travNode) && (travNode->node->level == node->level)) {
						neigbhourHoodGrid[n] = *(((VolGridNode *)travNode->node)->distanceValues[-m + 7]);
						break;
					}
				}
			}
		}

		for (n = 0; n < 8; n++) {

			invalidNodesCount = oppositeSignCount = 0;
			
			for (int k = 0; k < 27; k++) {
				
				if (k == 13) {
					continue;
				}

				if (neigbhourHoodGrid[distValueProjectionTable[n][k]] == FLT_MAX) {
					invalidNodesCount++;
				}
				else {
					if (neigbhourHoodGrid[distValueProjectionTable[n][13]] * neigbhourHoodGrid[distValueProjectionTable[n][k]] < 0) {
						oppositeSignCount++;
					}
				}
			}

			if ((invalidNodesCount > invalidNodes) && (oppositeSignCount > oppositeSign)) {
				(*i)->boundaryNode = true;
				this->boundaryNodeList.insert((*i));
				boundaryNodesCount++;
			}
		}
	}

	this->nofBoundaryNodes = boundaryNodesCount;

	qDebug("Boundary nodes: %i", this->nofBoundaryNodes);

	// init blending
	initBlendingBoundaryNodes();

	this->grid->nodeFactory->freeTraversalNode(travNode);

}


void VolGridNarrowBandIterator::clearBoundaryNodes() {

	for (std::set<class VolGridNode*>::iterator i = this->nodeList.begin(); i != this->nodeList.end(); i++) {
		if ((*i)->boundaryNode) {
			(*i)->boundaryNode = false;
			(*i)->original = true;
		}
	}
	this->boundaryNodeList.clear();
	this->diffusionNodeList.clear();
}


void VolGridNarrowBandIterator::initBlendingBoundaryNodes() {
	
	this->boundaryNodePos = new Vector3D[this->nofBoundaryNodes];

	int counter = 0;

	for (std::set<class VolGridNode*>::iterator i = this->nodeList.begin(); i != this->nodeList.end(); i++) {
		if (((*i)->boundaryNode) && (counter < this->nofBoundaryNodes)) {
			// get coordinates in [0,1]x[0,1]x[0,1]
			(*i)->getCoordinates(boundaryNodePos[counter], (OcTree *)this->grid);
			counter++;
		}
	}

	this->boundaryNodes = new NeighbourHood (this->boundaryNodePos, this->nofBoundaryNodes);
}


float VolGridNarrowBandIterator::getBlendingValue(VolGridNode *node) {
	
	Vector3D pos, boundaryNode;

	node->getCoordinates(pos, (OcTree *)this->grid);
	this->boundaryNodes->setSourcePoint(pos);
	this->boundaryNodes->setNofNeighbours(1);
	boundaryNode = this->boundaryNodes->getNeighbourPosition(0);

	float length = this->grid->length * 2;
	pos *= length;
	boundaryNode *= length;
	float distance = Vector3D::distance(pos, boundaryNode);
	float maxValue = this->largestHole*this->grid->lengthTable[this->grid->maxDepth];;
	if (distance >= maxValue) {
		return 0.0f;
	}
	else {
		float tmp = ((0.1*pow(distance/maxValue, 2))) - 1;
		return pow(tmp, 2);
	}
}


void VolGridNarrowBandIterator::smooth(int iteration, float influence) {

	int j, m, n;

	int distanceValueProjection[8];
	float neigbhourHoodGrid[64];
	VolGridNode *node;
	Vector3D point;
	VolGridNode **neighbours;
	VolGridTraversalNode *travNode =  new VolGridTraversalNode();
	
	for (int o = 0; o < iteration; o++) {
		std::set<class VolGridNode*>::iterator i;
		for (i = this->nodeList.begin(); i != this->nodeList.end(); i++) {

			for (n = 0; n < 64; n++) {
				neigbhourHoodGrid[n] = FLT_MAX;
			}

			neighbours = (VolGridNode **)(*i)->getNeighbourCellsSameLevel(this->grid);

			for (n = 0; n < 27; n++) {

				neighbourHoodInfo* nbrTable = &(neighbourHoodTable[n]);
				memcpy(distanceValueProjection, nbrTable->distanceValueProjection, 8 * sizeof(int));
					
				if (n < 13) {
					node = neighbours[n];
				}
				else if (n > 13) {
					node = neighbours[n - 1];
				}
				else {
					node = (*i);
				}

				if (node) {
					for (m = 0; m < 8; m++) {
						if (neigbhourHoodGrid[distanceValueProjection[m]] == FLT_MAX) {
							neigbhourHoodGrid[distanceValueProjection[m]] = *node->distanceValues[m];
						}
					}				
				}
			}

			delete []neighbours;
			neighbours = NULL;

			// fill nodes in the neighbourhood grid which are not filled in with direct neighbour information
			for (n = 0; n < 64; n++) {
				node = (*i);
				if (neigbhourHoodGrid[n] == FLT_MAX) {
					
					node->getCoordinates(point,this->grid);

					float length = this->grid->lengthTable[this->grid->maxDepth];
					point[0] = neighbourOffsets[n][0] * length;
					point[1] = neighbourOffsets[n][1] * length;
					point[2] = neighbourOffsets[n][2] * length;

					// look-up all eight possible neighbours
					// if there is a neighbour fill in the value and go to the next cell in the neigbourhood grid
					for (m = 0; m < 8; m++) {
						if (this->grid->getCellForPoint(point + octreeChildCenterOffset[m] * length, travNode) && (travNode->node->level == node->level)) {
							neigbhourHoodGrid[n] = *(((VolGridNode *)travNode->node)->distanceValues[-m + 7]);
							break;
						}
					}
				}
			}

			// calculate diffusion for every point
			for (j = 0; j < 8; j++) {

				int totalValues = 0;	// how many neighbours have a distance value
				float totalLowPassFilterSum = 0.0f;
				float totalSum = 0.0f;
				float distValue;

				for (int k = 0; k < 27; k++) {
					distValue = neigbhourHoodGrid[distValueProjectionTable[j][k]];
					
					if ((distValue != FLT_MAX) && (this->grid->filter[k] != 0)) {
						totalSum += distValue * this->grid->filter[k];
						totalValues++;
						totalLowPassFilterSum += this->grid->filter[k];
					}
				}

				if (totalValues > 0) {
					*(*i)->diffusionValues[j] = totalSum / totalLowPassFilterSum;
				}
			}
		}

		int counter;
		for (i = this->nodeList.begin(); i != this->nodeList.end(); i++) {
			counter = 0;
			for (int n = 0; n < 8; n++) {
				if (*(*i)->diffusionValues[n] != FLT_MAX) {
					*(*i)->distanceValues[n] = (influence * (*(*i)->diffusionValues[n])) + ((1 - influence) * (*(*i)->distanceValues[n]));
					counter++;
				}
			}
			if (counter == 8) {
				(*i)->valid = true;
			}
			else {
				(*i)->valid = false;
			}
		}
	}

	this->grid->nodeFactory->freeTraversalNode(travNode);

/*
	//////////////////////////// testing 
	int counter = 0;
	int ok = 0;
	for (i = this->nodeList.begin(); i != this->nodeList.end(); i++) {
		
		for (int n = 0; n < 8; n++) {

			//if (*(*i)->diffusionValues[n] != FLT_MAX) {
			//	(*i)->distanceValues[n] = *(*i)->diffusionValues[n];
			//}

			int	nbNode[27];
			int distValue[27];
			int nbCopyNode[7];
			int nbCopyNodePos[7];
			neighbourDistancePositions* nbrDisPos = &(neighbourDistanceTable[n]);
			
			memcpy(nbNode, nbrDisPos->neighbourNode, 27 * sizeof(int));
			memcpy(distValue, nbrDisPos->distanceValueNumber, 27 * sizeof(int));	
			memcpy(nbCopyNode, nbrDisPos->nbCopyNode, 7 * sizeof(int));
			memcpy(nbCopyNodePos, nbrDisPos->nbCopyNodePos, 7 * sizeof(int));

			for (int p = 0; p < 7; p++) {
				
				VolGridNode *copyNode = ((VolGridNode *)((*i)->neighbours[nbCopyNode[p]]));

				if (copyNode) {
					if ((*i)->distanceValues[n] != copyNode->distanceValues[nbCopyNodePos[p]]) {
						
						//delete copyNode->distanceValues[nbCopyNodePos[p]];
						copyNode->distanceValues[nbCopyNodePos[p]] = (*i)->distanceValues[n];
						counter++;
						ok++;
					}
					else {
						ok++;
					}
				}
			}
		}
	}
		qDebug("counter: %i/%i", counter, ok);

	//////////////////////////// testing 
*/
}


void VolGridNarrowBandIterator::diffuse(int iteration, bool blending, float blendingConstant) {


	int j, m, n;

	int distanceValueProjection[8];
	float neigbhourHoodGrid[64];
	VolGridNode *node;
	Vector3D point;
	VolGridNode **neighbours;
	VolGridTraversalNode *travNode = (VolGridTraversalNode *)this->grid->nodeFactory->newTraversalNode();

	for (int o = 0; o < iteration; o++) {
		std::set<class VolGridNode*>::iterator i;
		for (i = this->diffusionNodeList.begin(); i != this->diffusionNodeList.end(); i++) {	

			for (n = 0; n < 64; n++) {
				neigbhourHoodGrid[n] = FLT_MAX;
			}

			neighbours = (VolGridNode **)(*i)->getNeighbourCellsSameLevel(this->grid);

			for (n = 0; n < 27; n++) {

				neighbourHoodInfo* nbrTable = &(neighbourHoodTable[n]);
				memcpy(distanceValueProjection, nbrTable->distanceValueProjection, 8 * sizeof(int));
					
				if (n < 13) {
					node = neighbours[n];
				}
				else if (n > 13) {
					node = neighbours[n - 1];
				}
				else {
					node = (*i);
				}

				if (node) {
					for (m = 0; m < 8; m++) {
						if (neigbhourHoodGrid[distanceValueProjection[m]] == FLT_MAX) {
							neigbhourHoodGrid[distanceValueProjection[m]] = *node->distanceValues[m];
						}
					}			
				}
			}

			delete []neighbours;
			neighbours = NULL;

			// fill nodes in the neighbourhood grid which are not filled in with direct neighbour information
			for (n = 0; n < 64; n++) {
				node = (*i);
				if (neigbhourHoodGrid[n] == FLT_MAX) {
					
					node->getCoordinates(point,this->grid);

					float length = this->grid->lengthTable[this->grid->maxDepth];
					point[0] = neighbourOffsets[n][0] * length;
					point[1] = neighbourOffsets[n][1] * length;
					point[2] = neighbourOffsets[n][2] * length;

					// look-up all eight possible neighbours
					// if there is a neighbour fill in the value and go to the next cell in the neigbourhood grid
					for (m = 0; m < 8; m++) {
						if (this->grid->getCellForPoint(point + octreeChildCenterOffset[m] * length, travNode) && (travNode->node->level == node->level)) {
							neigbhourHoodGrid[n] = *(((VolGridNode *)travNode->node)->distanceValues[-m + 7]);
							break;
						}
					}
				}
			}

			// calculate diffusion for every point
			for (j = 0; j < 8; j++) {

				int totalValues = 0;	// how many neighbours have a distance value
				float totalLowPassFilterSum = 0.0f;
				float totalSum = 0.0f;
				float distValue;

				for (int k = 0; k < 27; k++) {
					distValue = neigbhourHoodGrid[distValueProjectionTable[j][k]];
					
					if ((distValue != FLT_MAX) && (this->grid->filter[k] != 0)) {
						totalSum += distValue * this->grid->filter[k];
						totalValues++;
						totalLowPassFilterSum += this->grid->filter[k];
					}
				}

				if (totalValues > 0) {
					*(*i)->diffusionValues[j] = totalSum / totalLowPassFilterSum;
				}
			}
		}


		int counter;

		for (i = this->diffusionNodeList.begin(); i != this->diffusionNodeList.end(); i++) {
			counter = 0;
			for (int n = 0; n < 8; n++) {
				if (*(*i)->diffusionValues[n] != FLT_MAX) {
					// with blending
					if (blending) {
						float blendValue = getBlendingValue(*i);
						if ((*(*i)->distanceValues[n] != FLT_MAX) && ((*i)->original)) {
							blendValue *= blendingConstant;
							*(*i)->distanceValues[n] = blendValue*(*(*i)->diffusionValues[n]) + (1-blendValue)*(*(*i)->distanceValues[n]);
						}
						else {
							*(*i)->distanceValues[n] = (*(*i)->diffusionValues[n]);
						}
					}
					// no blending
					else {
					*(*i)->distanceValues[n] = (*(*i)->diffusionValues[n]);
					}
				}
				counter++;
			}

			if (counter == 8) {
				(*i)->valid = true;
			}
		}
	}
	
	this->grid->nodeFactory->freeTraversalNode(travNode);
}


bool VolGridNarrowBandIterator::iterate(OcTreeTraversalNode *ocTravNode) {

    VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

	// determine if node is a on level above leaf node
	if (travNode->depth == this->grid->getMaxDepth() && ((VolGridNode *) (travNode->node))->nodeType == LEAF) {
		this->nodeList.insert(node);
		((VolGridNode *)travNode->node)->original = true;
	}

    return true;
}


void VolGridNarrowBandIterator::getNarrowBand() {
		
	this->grid->traverse(this);
	qDebug("Number of Voxels in narrow band: %i", this->nodeList.size());
}


bool VolGridActiveCellCounter::iterate(OcTreeTraversalNode *ocTravNode) {

	VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

	// Determine if child is a leaf node at maximum depth.
	if (travNode->depth == this->grid->getMaxDepth() && ((VolGridNode *) (travNode->node))->nodeType == LEAF) {

		// Decide whether cell is active or not.
		int type = 0;
		if ((*(node->distanceValues[0])) > 0.0f) type |=   1;
		if ((*(node->distanceValues[1])) > 0.0f) type |=   2;
		if ((*(node->distanceValues[2])) > 0.0f) type |=   4;
		if ((*(node->distanceValues[3])) > 0.0f) type |=   8;
		if ((*(node->distanceValues[4])) > 0.0f) type |=  16;
		if ((*(node->distanceValues[5])) > 0.0f) type |=  32;
		if ((*(node->distanceValues[6])) > 0.0f) type |=  64;
		if ((*(node->distanceValues[7])) > 0.0f) type |= 128;

		if (type == 0 || type == 255) {
			return true;
		}
		
		// cell is active since not all values are positive or negative
		this->nofActiveCells++;
	}

	return true;
}


void VolGridActiveCellCounter::count() {

	this->nofActiveCells = 0;
	this->grid->traverse(this);
}


bool VolGridVariation::iterate(OcTreeTraversalNode *ocTravNode) {

	VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

	// Determine if child is a leaf node at maximum depth.
	if ((travNode->depth == this->grid->getMaxDepth()) && (((VolGridNode *) (travNode->node))->nodeType == LEAF) &&
		(((VolGridNode *) (travNode->node))->valid == true)) {
		
		// Compute whether current cell is active.
		int type = 0;
		if (*(node->distanceValues[0]) > 0.0f) type |=   1;
		if (*(node->distanceValues[1]) > 0.0f) type |=   2;
		if (*(node->distanceValues[2]) > 0.0f) type |=   4;
		if (*(node->distanceValues[3]) > 0.0f) type |=   8;
		if (*(node->distanceValues[4]) > 0.0f) type |=  16;
		if (*(node->distanceValues[5]) > 0.0f) type |=  32;
		if (*(node->distanceValues[6]) > 0.0f) type |=  64;
		if (*(node->distanceValues[7]) > 0.0f) type |= 128;

		// Return if there is no active cell.
		if (type == 0 || type == 255) {
			return true;
		}

		// Compute variation for every active cell.
		this->variation[cellNumber] = this->grid->computeSurfaceVariation(travNode->center);
		this->variationAvg += this->variation[cellNumber];
		this->cellNumber++;
	}

	return true;
}


void VolGridVariation::calculateVariation() {

	this->cellNumber = 0;
	this->variationAvg = 0.0f;
	this->grid->traverse(this);
	if (this->nofActiveCells > 0) {
		this->variationAvg /= this->nofActiveCells;
	}
}


Vector3D VolGrid::estimateGradient(Vector3D relativePos, VolGridNode* node, float cellLength) {

	Vector3D gradient;

	float h[6];

	// p, q, r are distances in x, y, z direction
	// trilinear interpolation
	// g0 = (1-p/a)*f[0] + (p/a) * f[1];
	// g1 = (1-p/a)*f[2] + (p/a) * f[3];
	// g2 = (1-p/a)*f[4] + (p/a) * f[5];
	// g3 = (1-p/a)*f[6] + (p/a) * f[7];
	// h0 = (1-q/a)*g0 + (q/a) * g1;
	// h1 = (1-q/a)*g2 + (q/a) * g3;
	// value = (1-(r/a))*h0 + (r/a)*h1;

	// Gradient estimation based on trilinear interpolation and central differences.
	float a = cellLength;
	float p, q, r;
	float g0, g1, g2, g3;
	q = relativePos[1];
	r = relativePos[2];

	float h0 = (1-q/a)*(*(node->distanceValues[0])) + (q/a) * (*(node->distanceValues[2]));
	float h1 = (1-q/a)*(*(node->distanceValues[4])) + (q/a) * (*(node->distanceValues[6]));
	h[0] = (1-(r/a))*h0 + (r/a)*h1;

	p = cellLength;
	q = relativePos[1];
	r = relativePos[2];
	h0 = (1-q/a)*(*(node->distanceValues[1])) + (q/a) * (*(node->distanceValues[3]));
	h1 = (1-q/a)*(*(node->distanceValues[5])) + (q/a) * (*(node->distanceValues[7]));
	h[1] = (1-r/a)*h0 + (r/a)*h1;
	
	p = relativePos[0];
	r = relativePos[2];
	g0 = (1-p/a)*(*(node->distanceValues[0])) + (p/a) * (*(node->distanceValues[1]));
	g2 = (1-p/a)*(*(node->distanceValues[4])) + (p/a) * (*(node->distanceValues[5]));
	h[2] = (1-(r/a))*g0 + (r/a)*g2;

	q = cellLength;
	g1 = (1-p/a)*(*(node->distanceValues[2])) + (p/a) * (*(node->distanceValues[3]));
	g3 = (1-p/a)*(*(node->distanceValues[6])) + (p/a) * (*(node->distanceValues[7]));
	h[3] = (1-r/a)*g1 + (r/a)*g3;

	p = relativePos[0];
	q = relativePos[1];
	g0 = (1-p/a)*(*(node->distanceValues[0])) + (p/a) * (*(node->distanceValues[1]));
	g1 = (1-p/a)*(*(node->distanceValues[2])) + (p/a) * (*(node->distanceValues[3]));
	h[4] = (1-q/a)*g0 + (q/a) * g1;

	r = cellLength;
	g2 = (1-p/a)*(*(node->distanceValues[4]))+ (p/a) * (*(node->distanceValues[5]));
	g3 = (1-p/a)*(*(node->distanceValues[6])) + (p/a) * (*(node->distanceValues[7]));
	h[5] = (1-q/a)*g2 + (q/a) * g3;

	// central differences
	gradient[0] = h[1] - h[0];
	gradient[1] = h[3] - h[2];
	gradient[2] = h[5] - h[4];
	
	gradient.normalize();
	return gradient;
}

/*
// other method (more accurate, but slower)
Vector3D VolGrid::estimateGradient(Vector3D relativePos, VolGridNode* node, float cellLength) {

	Vector3D gradient[8];
	int n, m;

	int distanceValueProjection[8];
	float neigbhourHoodGrid[64];
	VolGridNode *Nnode;
	VolGridNode **neighbours;
	Vector3D point;

	VolGridTraversalNode *travNode = (VolGridTraversalNode *)this->nodeFactory->newTraversalNode();

		for (n = 0; n < 64; n++) {
			neigbhourHoodGrid[n] = FLT_MAX;
		}

		neighbours = (VolGridNode **)node->getNeighbourCells((OcTree*) this);
		
		for (n = 0; n < 27; n++) {

			neighbourHoodInfo* nbrTable = &(neighbourHoodTable[n]);
			memcpy(distanceValueProjection, nbrTable->distanceValueProjection, 8 * sizeof(int));
				
			if (n < 13) {
				Nnode = neighbours[n];
			}
			else if (n > 13) {
				Nnode = neighbours[n - 1];
			}
			else {
				Nnode = (node);
			}

			if (Nnode) {
				for (m = 0; m < 8; m++) {
					if (neigbhourHoodGrid[distanceValueProjection[m]] == FLT_MAX) {
						neigbhourHoodGrid[distanceValueProjection[m]] = *Nnode->distanceValues[m];
					}
				}				
			}
		}

		delete []neighbours;
		neighbours = NULL;

		// fill nodes in the neighbourhood grid which are not filled in with direct neighbour information
		for (n = 0; n < 64; n++) {

			if (neigbhourHoodGrid[n] == FLT_MAX) {
				
				node->getCoordinates(point,this);

				float length = this->lengthTable[this->maxDepth];
				point[0] = neighbourOffsets[n][0] * length;
				point[1] = neighbourOffsets[n][1] * length;
				point[2] = neighbourOffsets[n][2] * length;
				
				// look-up all eight possible neighbours
				// if there is a neighbour fill in the value and go to the next cell in the neigbourhood grid
				for (m = 0; m < 8; m++) {
					if (this->getCellForPoint(point + octreeChildCenterOffset[m] * length, travNode) && (travNode->node->level == node->level)) {
						neigbhourHoodGrid[n] = *(((VolGridNode *)travNode->node)->distanceValues[-m + 7]);
						break;
					}
				}
			}
		}

	gradient[0][0] = 0.5 * (neigbhourHoodGrid[22] - neigbhourHoodGrid[20]);
	gradient[0][1] = 0.5 * (neigbhourHoodGrid[25] - neigbhourHoodGrid[17]);
	gradient[0][2] = 0.5 * (neigbhourHoodGrid[37] - neigbhourHoodGrid[5]);
	gradient[1][0] = 0.5 * (neigbhourHoodGrid[23] - neigbhourHoodGrid[21]);
	gradient[1][1] = 0.5 * (neigbhourHoodGrid[26] - neigbhourHoodGrid[18]);
	gradient[1][2] = 0.5 * (neigbhourHoodGrid[38] - neigbhourHoodGrid[6]);
	gradient[2][0] = 0.5 * (neigbhourHoodGrid[26] - neigbhourHoodGrid[24]);
	gradient[2][1] = 0.5 * (neigbhourHoodGrid[29] - neigbhourHoodGrid[21]);
	gradient[2][2] = 0.5 * (neigbhourHoodGrid[41] - neigbhourHoodGrid[9]);
	gradient[3][0] = 0.5 * (neigbhourHoodGrid[27] - neigbhourHoodGrid[25]);
	gradient[3][1] = 0.5 * (neigbhourHoodGrid[30] - neigbhourHoodGrid[22]);
	gradient[3][2] = 0.5 * (neigbhourHoodGrid[42] - neigbhourHoodGrid[10]);
	gradient[4][0] = 0.5 * (neigbhourHoodGrid[38] - neigbhourHoodGrid[36]);
	gradient[4][1] = 0.5 * (neigbhourHoodGrid[41] - neigbhourHoodGrid[33]);
	gradient[4][2] = 0.5 * (neigbhourHoodGrid[53] - neigbhourHoodGrid[21]);
	gradient[5][0] = 0.5 * (neigbhourHoodGrid[39] - neigbhourHoodGrid[37]);
	gradient[5][1] = 0.5 * (neigbhourHoodGrid[42] - neigbhourHoodGrid[34]);
	gradient[5][2] = 0.5 * (neigbhourHoodGrid[54] - neigbhourHoodGrid[22]);
	gradient[6][0] = 0.5 * (neigbhourHoodGrid[42] - neigbhourHoodGrid[40]);
	gradient[6][1] = 0.5 * (neigbhourHoodGrid[45] - neigbhourHoodGrid[37]);
	gradient[6][2] = 0.5 * (neigbhourHoodGrid[57] - neigbhourHoodGrid[25]);
	gradient[7][0] = 0.5 * (neigbhourHoodGrid[43] - neigbhourHoodGrid[41]);
	gradient[7][1] = 0.5 * (neigbhourHoodGrid[46] - neigbhourHoodGrid[38]);
	gradient[7][2] = 0.5 * (neigbhourHoodGrid[58] - neigbhourHoodGrid[26]);


	// Gradient estimation based on trilinear interpolation and central differences.
	float a = cellLength;
	float p, q, r;
	Vector3D g0, g1, g2, g3, h0, h1;
	p = relativePos[0];
	q = relativePos[1];
	r = relativePos[2];

	g0 = (1-p/a)*gradient[0] + (p/a)*gradient[1];
	g1 = (1-p/a)*gradient[2] + (p/a)*gradient[3];
	g2 = (1-p/a)*gradient[4] + (p/a)*gradient[5];
	g3 = (1-p/a)*gradient[6] + (p/a)*gradient[7];
	h0 = (1-q/a)*g0 + (q/a)*g1;
	h1 = (1-q/a)*g2 + (q/a)*g3;
	Vector3D grad = (1-r/a)*h0 + (r/a)*h1;
	
	grad.normalize();
	return grad;
}
*/

float VolGrid::computeSurfaceVariation(Vector3D pos) {

	Vector3D gradient, adjacentGradient;
	Vector3D posInCell;
	VolGridNode *neighbourNode;
	float variation = 0.0;
	float cellLength;
	int counter = 0;

	VolGridTraversalNode *travNode = (VolGridTraversalNode *)this->nodeFactory->newTraversalNode();
	
	// calculate position normalized to [0, 1]
	Vector3D boundingBox = Vector3D (2 * this->length, 2 * this->length, 2 * this->length);
	Vector3D bBoxMinSquare = this->center - Vector3D(this->length, this->length, this->length);
	Vector3D normalizedPos = (pos - bBoxMinSquare) / (boundingBox);

	// Descend to octree cell.
	if (this->getCellForPoint(normalizedPos, travNode)) {

		cellLength = this->lengthTable[travNode->node->level - 1];
		posInCell = Vector3D(0.5 * cellLength, 0.5 * cellLength, 0.5 * cellLength);
		
		gradient = this->estimateGradient(posInCell, (VolGridNode *)travNode->node, cellLength);

		// Calculate the distance field variation with the gradient angles.
		// Zero means no variation in distance field.
		neighbourNode = (VolGridNode *)travNode->node->getNeighbourLeftSameLevel((OcTree *)this);
		if ((neighbourNode != NULL) && (neighbourNode->valid)) {
			adjacentGradient = this->estimateGradient(posInCell, neighbourNode, cellLength);
			variation += 1 - fabs(Vector3D::dotProduct(gradient, adjacentGradient));
			counter++;
		}
		neighbourNode = (VolGridNode *)travNode->node->getNeighbourRightSameLevel((OcTree *)this);
		if ((neighbourNode != NULL) && (neighbourNode->valid)) {
			adjacentGradient = this->estimateGradient(posInCell, neighbourNode, cellLength);
			variation += 1 - fabs(Vector3D::dotProduct(gradient, adjacentGradient));
			counter++;
		}
		neighbourNode = (VolGridNode *)travNode->node->getNeighbourFrontSameLevel((OcTree *)this);
		if ((neighbourNode != NULL) && (neighbourNode->valid)) {
			adjacentGradient = this->estimateGradient(posInCell, neighbourNode, cellLength);
			variation += 1 - fabs(Vector3D::dotProduct(gradient, adjacentGradient));
			counter++;
		}
		neighbourNode = (VolGridNode *)travNode->node->getNeighbourRearSameLevel((OcTree *)this);
		if ((neighbourNode != NULL) && (neighbourNode->valid)) {
			adjacentGradient = this->estimateGradient(posInCell, neighbourNode, cellLength);
			variation += 1 - fabs(Vector3D::dotProduct(gradient, adjacentGradient));
			counter++;
		}
		neighbourNode = (VolGridNode *)travNode->node->getNeighbourTopSameLevel((OcTree *)this);
		if ((neighbourNode != NULL) && (neighbourNode->valid)) {
			adjacentGradient = this->estimateGradient(posInCell, neighbourNode, cellLength);
			variation += 1 - fabs(Vector3D::dotProduct(gradient, adjacentGradient));
			counter++;
		}
		neighbourNode = (VolGridNode *)travNode->node->getNeighbourBottomSameLevel((OcTree *)this);
		if ((neighbourNode != NULL) && (neighbourNode->valid)) {
			adjacentGradient = this->estimateGradient(posInCell, neighbourNode, cellLength);
			variation += 1 - fabs(Vector3D::dotProduct(gradient, adjacentGradient));
			counter++;
		}

		// Compute variation.
		if (counter > 0) {
			variation /= counter;
		}
	}

	this->nodeFactory->freeTraversalNode(travNode);

	return variation;
}


bool VolGrid::projectPointOntoDistanceField(Vector3D &pos, Vector3D &gradient) {
	
	bool ok;
	float f[8];		
	int j;
	Vector3D cellPosLow;
	Vector3D posInCell;

	VolGridTraversalNode *travNode = (VolGridTraversalNode *)this->nodeFactory->newTraversalNode();
	Vector3D boundingBox = Vector3D (2 * this->length, 2 * this->length, 2 * this->length);
	Vector3D bBoxMinSquare = this->center - Vector3D(this->length, this->length, this->length);
	Vector3D normalizedPos = (pos - bBoxMinSquare) / (boundingBox);

	if (this->getCellForPoint(normalizedPos, travNode)) {

		if ((this->rootLevel - travNode->node->level) != this->maxDepth) {
			this->nodeFactory->freeTraversalNode(travNode);
			return false;
		}

		for (j = 0; j < 8; j++) {
			f[j] =  *(((VolGridNode *)travNode->node)->distanceValues[j]);
		}

		int type = 0;
		if (f[0] > 0.0) type |=   1;
		if (f[1] > 0.0) type |=   2;
		if (f[2] > 0.0) type |=   4;
		if (f[3] > 0.0) type |=   8;
		if (f[4] > 0.0) type |=  16;
		if (f[5] > 0.0) type |=  32;
		if (f[6] > 0.0) type |=  64;
		if (f[7] > 0.0) type |= 128;

		cellPosLow = pos - this->boundingBoxMin;
		float cellLength = this->lengthTable[travNode->node->level - 1];
		posInCell = Vector3D(fmod(cellPosLow[0], cellLength),  fmod(cellPosLow[1], cellLength), fmod(cellPosLow[2], cellLength));
		posInCell /= cellLength;

		float distValue = f[0]*(1-posInCell[0])*(1-posInCell[1])*(1-posInCell[2]) +
						  f[1]*(posInCell[0])*(1-posInCell[1])*(1-posInCell[2]) +
						  f[2]*(1-posInCell[0])*(posInCell[1])*(1-posInCell[2]) +
						  f[3]*(posInCell[0])*(posInCell[1])*(1-posInCell[2]) +
						  f[4]*(1-posInCell[0])*(1-posInCell[1])*(posInCell[2]) +
						  f[5]*(posInCell[0])*(1-posInCell[1])*(posInCell[2]) +
						  f[6]*(1-posInCell[0])*(posInCell[1])*(posInCell[2]) +
						  f[7]*(posInCell[0])*(posInCell[1])*(posInCell[2]);


		gradient = this->estimateGradient(posInCell, (VolGridNode *)travNode->node, cellLength);
		Vector3D translation = -gradient * distValue;
		pos += translation;
	}
	else {
		ok = false;
	}

	this->nodeFactory->freeTraversalNode(travNode);
	return ok;
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
