// Title:   VolGrid.h
// Created: Thu Sep 25 15:10:13 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __VOLGRID_H_
#define __VOLGRID_H_

#include "MovingLeastSquares.h"
#include "OcTree.h"
#include "NeighbourTable.h"
#include "myDefines.h"
#include "../../../DataTypes/src/Vector2D.h"
#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../Utilities/MagicSoft/MgcMath.h"
#include "../../../Utilities/MagicSoft/MgcMatrix2.h"
#include "../../../Utilities/MagicSoft/MgcEigen.h"
#include "../../../Utilities/MagicSoft/MgcAppr3DPlaneFit.h"
#include "../../../Utilities/MagicSoft/MgcVector3.h"
#include "../../../Utilities/src/StringTools.h"
#include <vector>
#include <set>
#include <map>
#include <qgl.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_linalg.h>


/**
 * VolGrid node types
 */
enum VOLGRID_NODE {
	INTERNAL,	// node contains narrow band - but it is not a leaf node
	EXTERNAL,   // node does not contain the narrow band (it is a OcTree leaf node)
	LEAF,		// narrow band
};


/**
 * A VolGrid node contains information about the node type and a signed
 * distance from the node's vertices to its surface.
 */
class VolGridNode : public OcTreeNode {

public:
	VolGridNode() { original = false;};
	
	~VolGridNode() {
		for (int i = 0; i < 8; i++) {
			if (this->distanceValues[i] != NULL) {
				delete this->distanceValues[i];
			}
			if (this->diffusionValues[i] != NULL) {
				delete this->diffusionValues[i];
			}
		}
	}

public:
	/// The eight distance values (from lower left to upper right)
	float			*distanceValues[8];
	/// Indicates value of diffusion process
	float			*diffusionValues[8];
	/// Defines the node type (INTERNAL, EXTERNAL, LEAF)
    VOLGRID_NODE	nodeType;
	/// Valid tag
	bool			valid;
	/// Original distance field tag
	bool			original;
	/// Boundary node tag
	bool			boundaryNode;

};

/**
 * A volumetric grid based on the <code>OcTree</code> base class.
 */
class VolGrid : public OcTree {

friend class VolGridBuilder;
friend class VolGridReader;

public:

	VolGrid() {};
	
	~VolGrid() {
		delete (VolGridNode *)this->root;
		delete this->slice;
		delete this->boundary;
		delete this->nodeFactory;
	};

	/**
	 * Creates a volumetric grid.
	 * @param bBoxMin
	 *				lower left bounding box coordinate.
	 * @param bBoxMax
	 *				upper right bounding box coordinate.
	 * @param maxDepth
	 *				defines the maximum depth of the volumetric grid.
	 * @param narrowBandNofVoxels
	 *				defines the width of the narrow band with number of voxels.
	 */
	void createVolGrid(Vector3D& bBoxMin, Vector3D& bBoxMax, float bBoxSize, int maxDepth, int narrowBandNofVoxels);
    
	/**
 	 * @param level
	 *				specifies the level.
	 * @param sliceNumber
	 *				specifies the desired slice number.
	 * @param plane
	 *				specifies the desired plane.
	 * @return
	 *				returns true if all parameters are set correctly.
	 */
	bool getSlice(int level, int sliceNumber, PLANE plane);

	/**
     * Returns the side length of a slice at a specific level.
 	 * @param level
	 *				specifies the level
	 * @return
	 *				returns the number of voxels of a side length.
	 */
	int getSliceDimension(int level) { return pow(2, level); }

	/**
     * @return
	 *				the first entry of the slice values.
	 */
	float *getSlice() { return slice; }

	/**
     * @return
	 *				the maximal depth.
	 */
	int	getMaxDepth() { return maxDepth; }
	
	/**
     * Calculates the diamter of the narrow band for a level.
	 * @param narrowBandNofVoxels
	 *				the number of voxels for the narrow band.
	 * @return
	 *				returns the width of the narrow band.
	 */
	float calculateNarrowBandWidth(int narrowBandNofVoxels);

	/**
     * @return
	 *				the narrow band width.
	 */
	float getNarrowBandWidth() { return narrowBandWidth; }

	/**
     * Projects a point to the distance field
	 */
	bool projectPointOntoDistanceField(Vector3D &pos, Vector3D &gradient);
	
	/**
     * Estimates the gradient with central differences and trilinear interpolation
	 */
	Vector3D estimateGradient(Vector3D relativePos, VolGridNode* node, float cellLength);
	
	/**
     * Approximates Surface Variation
	 */
	float computeSurfaceVariation(Vector3D pos);


public:
	/// Precalculated length table.
	float 					lengthTable[16];
	/// Specifies the maximum depth to iterate.
	int						maxDepth;
	float					*boundary;
	/// Values of <code>getSlice</code>
	float					*slice;
	/// Lower left bounding box coordinate.
	Vector3D				boundingBoxMin;
	/// Upper right bounding box coordinate.
	Vector3D				boundingBoxMax;
	// distance field calculation method
	DISTANCEFIELDCALC		distanceFieldCalcMethod;
	/// number of neighbours
	int						distanceFieldNofNeighbours;
	/// Filter for diffusion
	int						filter[27];
protected:
	/// Defines the width of the narrow band.
	float 					narrowBandWidth;
	int						nofVoxels;
	/// Number of neighbours used when calculating distance to surface
	int						nofNeighbours;

};


// A marching Cubes entry
struct MCEntry {
    int nVertices;
    int edgeStart[12];
    int edgeEnd[12];
    int nTriangles;
    int triangles[5][3];
};


// load the modified Marching Cubes look-up table
static MCEntry table[256] = {
#include "MCTable.h"
};


/**
 * A VolGrid traversal node
 */
class VolGridTraversalNode : public OcTreeTraversalNode {

};

/**
 * VolGridBuilder is used to generate a fully populated OcTree with
 * a signed distance function.
 */
class VolGridBuilder : public OcTreeIterator {

public:
  /** 
	 * @param grid
	 *				reference to the volumetric grid.	
	* <b>Note:</b> showInternalNodes is set to false.
	*/
	VolGridBuilder(VolGrid *grid) {
		showInternalNodes = false;
		this->grid = grid;
	}
	
	/**
	 * Destructor deletes surfelPositions and surfelNormals.
	 */
	~VolGridBuilder() {
		if (surfelPositions != NULL)
            delete[] surfelPositions;
        if (surfelNormals != NULL)
            delete[] surfelNormals;
		delete search;
	}
	
	/**
	 * Builds a populated <code>OcTree</code>.
	 */
    void build();

	/**
	 * Needed to iterate the <code>OcTree</code>.
	 */	
	virtual bool iterate(OcTreeTraversalNode *travNode);

protected:
	/**
	 * Calculates the distance to the next point from the center
	 * @param travNode
	 *					is the traversal node.
	 */
	bool getDistanceFromCenter(VolGridTraversalNode* travNode);
	
	
	// different distance field calculation methods
	float getDistanceNextPoint(VolGridTraversalNode *travNode, Vector3D queryPos);
	float getDistanceTangentPlane(VolGridTraversalNode *travNode, Vector3D queryPos, int neighbourHoodSize);
	float getDistanceWeightedLeastSquares(VolGridTraversalNode *travNode, Vector3D queryPos,  int neighbourHoodSize);
	float getDistanceOptWeightedLeastSquares(VolGridTraversalNode *travNode, Vector3D queryPos, int neighbourHoodSize);

	/**
	 * Calculate all distance values at the corners of a cell.
	 * @param travNode
	 *					is the traversal node.
	 */
	void getDistanceValues(VolGridTraversalNode* travNode);

	/**
	 * Returns true whenever a VolGridNode should be splitted up.
     * @param travNode
	 *					is the traversal node.
	 */
    bool shouldSplit(VolGridTraversalNode *travNode);

	/**
	 * Initialize nearest neighbours (kd-tree)
	 */
	void initNearestNeighbours();

public:

	/// the volumetric grid	
	VolGrid		  *grid;
	// vector to store the surfel positions (used for nearest neighbours with kd-tree)
	Vector3D	  *surfelPositions;
	// vector to store the surfel normals (used for nearest neighbours with kd-tree)
	Vector3D	  *surfelNormals;
	// neighbour hood  (used for nearest neighbours with kd-tree)
	NeighbourHood *search;
	// number of surfels
	int			  nofSurfels;

};




/**
 * VolGridNarrowBandIterator is used to traverse over the narrow band.
 */
class VolGridNarrowBandIterator : public OcTreeIterator {

public:
	/**
	 * Constructor.
	 */	
	VolGridNarrowBandIterator(VolGrid *grid) {
		showInternalNodes = true;
		this->grid = grid;

	}
	
	/**
	 * Destructor.
	 */	
	~VolGridNarrowBandIterator() {
		nodeList.clear();
		boundaryNodeList.clear();
		diffusionNodeList.clear();
		//delete this->boundaryNodes;
		//delete []this->boundaryNodePos;
	}

	/**
	 * Iterator.
	 */	
	virtual bool iterate(OcTreeTraversalNode *travNode);

	/**
	 * Smooth voxels.
	 */	
	void smooth(int iteration, float influence);

	/**
	 * Smooth voxels.
	 */	
	void diffuse(int iteration, bool blending, float blendingConstant);
	
	/**
	 * Find boundary nodes and mark them.
	 */
	void findBoundaryNodes(int oppositeSign, int invalidNodes);

	/**
	 * Clear all boundary nodes
	 */
	void clearBoundaryNodes();

	/**
	 * Grow the OcTree by value in all directions (cubical enlargement)
	 */	
	void growVolGrid(int value);

	/**
	 * Returns the narrow band.
	 */	
	void getNarrowBand();


	VolGridNode* createVolGridNode(VolGridNode *node, unsigned int nextLevel, unsigned int xLocCode, unsigned int yLocCode, unsigned int zLocCode);


	void initBlendingBoundaryNodes();
	float getBlendingValue(VolGridNode *node);


public:

	int						largestHole;
protected: 
	/// Reference to the volumetric grid.
	VolGrid					*grid;
	/// A vector with all nodes from the narrow band.
	std::set<VolGridNode *>	nodeList;
	/// all boundary nodes
	std::set<VolGridNode *>	boundaryNodeList;
	/// diffusion volume
	std::set<VolGridNode *>	diffusionNodeList;
	int						nofBoundaryNodes;
	NeighbourHood			*boundaryNodes;
	Vector3D				*boundaryNodePos;
	float					diffusionSmoothAlpha;
};


/**
 * VolGridActiveCellCounter is used to count the cells with sign changes.
 */
class VolGridActiveCellCounter : public OcTreeIterator {

public:
	/**
	 * Constructor.
	 */	
	VolGridActiveCellCounter(VolGrid *grid) { showInternalNodes = false; this->grid = grid; }
	
	/**
	 * Destructor.
	 */	
	~VolGridActiveCellCounter() { }

	/**
	 * Iterator.
	 */	
	virtual bool iterate(OcTreeTraversalNode *travNode);

	/**
	 * Build is called to generate the iso-surface.
	 */	
	void count();

public:
	VolGrid	*grid;
	int nofActiveCells;
};



/**
 * VolGridActiveCellCounter is used to count the cells with sign changes.
 */
class VolGridVariation : public OcTreeIterator {

public:
	/**
	 * @param grid
	 *				reference to the volumetric grid.
	 * @param nofActiveCells
	 *				set number of active cells.
	 */
	VolGridVariation(VolGrid *grid, int nofActiveCells) {
		showInternalNodes = false;
		this->grid = grid;
		this->nofActiveCells = nofActiveCells;
		this->variation = new float [nofActiveCells];
	}
	
	~VolGridVariation() { }

	/**
	 * Needed to iterate the octree.
	 */	
	virtual bool iterate(OcTreeTraversalNode *travNode);

	/**
	 * Calculates the variation.
	 */	
	void calculateVariation();

public:
	/// number of cells
	int		cellNumber;
	/// reference to the grid
	VolGrid	*grid;
	/// number of cells which contain the surface
	int		nofActiveCells;
	/// variation
	float	*variation;
	/// average variation
	float	variationAvg;
};

#endif	// __VOLGRID_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
