// Title:   VolGridEllipticalMC.h
// Created: Thu Sep 25 15:10:09 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __VOLGRIDELLIPTICALMARCHINGCUBES_H_
#define __VOLGRIDELLIPTICALMARCHINGCUBES_H_

#include "VolGrid.h"

/**
 * <i>Elliptical Marching Cubes</i> algorithm is a adaptation of the <i>Marching Cubes</i>
 * algorihm for elliptical surfels.
 */
class VolGridEllipticalMarchingCubes : public OcTreeIterator {

public:
	
	/** 
	 * @param grid
	 *				reference to the volumetric grid.
	 */
	VolGridEllipticalMarchingCubes(VolGrid *grid) {
		showInternalNodes = false;
		this->grid = grid;	
	}
	
	~VolGridEllipticalMarchingCubes() { }

	/**
	 * Needed to iterate the octree.
	 */	
	virtual bool iterate(OcTreeTraversalNode *travNode);

    /**
	 * This method determines the triangle positions of a volumetric grid cell according
	 * to the modified <i>Marching Cubes</i> algorithm.
	 *
	 * @param values[8]
	 *					the distance values from the eight corners of the cube.
	 * @param level
	 *					defines the iso-level to extract from (normally level = 0.0)
	 * @param nVertices[8]
	 *					defines the number of vertices according to the <i>Marching Cubes</i> look-up table.
	 * @param edgeStart[12]
	 *					defines on which edge the corresponding edge start is according to the <i>Marching Cubes</i> look-up table.
	 * @param edgeEnd[12]
	 *					defines on which edge the corresponding edge end is according to the <i>Marching Cubes</i> look-up table.
	 * @param param[12]
	 *					defines the where the point has to be located [0,1] on an edge.
	 * @param nTriangles
	 *					number of triangles according to the <i>Marching Cubes</i> look-up table.
	 * @param triangles[5][3]
	 *					triangles according to the <i>Marching Cubes</i> look-up table.
	 */
	void marchingCubes(float values[8], float level, int& nVertices, int edgeStart[12], int edgeEnd[12],
		               float param[12], int& nTriangles, int triangles[5][3]);

	/**
	 * This method is called to run the <i>Elliptical Marching Cubes</i> algorithm.
	 * 
	 * @param isovalue
	 *					defines the level to extract the surface from.
	 */	
	void build(float isovalue);

	/**
	 * This method is called to generate the surfels.
	 */	
	void generateSurface();

public:
	/// Reference to the volumetric grid.
	VolGrid						  *grid;
	/// Surfel positions
	std::vector<class Vector3D>   posList;	
	/// Surfel normals
	std::vector<class Vector3D>   normalList;
	/// Tangent axes
	std::vector<class Vector3D>   tangentAxesList;
	/// Iso-value to extract from
	float						  isoValue;
};

#endif	// __VOLGRIDELLIPTICALMARCHINGCUBES_H_
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
