// Title:   VolGridMesh.cpp
// Created: Thu Sep 25 15:10:07 2003
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "VolGridMesh.h"

void VolGridMesh::marchingCubes(float values[8], float level, int& nVertices, int edgeStart[12],
								int edgeEnd[12], float param[12], int& nTriangles, int triangles[5][3]) {

	// Compute the "case" number
    int type = 0;
    if (values[0] > level) type |=   1;
    if (values[1] > level) type |=   2;
    if (values[2] > level) type |=   4;
    if (values[3] > level) type |=   8;
    if (values[4] > level) type |=  16;
    if (values[5] > level) type |=  32;
    if (values[6] > level) type |=  64;
    if (values[7] > level) type |= 128;

    // Return if nothing to do
    if (type == 0 || type == 255) {
		nVertices = 0;
		nTriangles = 0;
		return;
    }

    // Get info about this case
    MCEntry *info = &(table[type]);

    // Copy info to output arguments
    nVertices  = info->nVertices;
    memcpy(edgeStart, info->edgeStart, 12 * sizeof(int));
    memcpy(edgeEnd, info->edgeEnd, 12 * sizeof(int));
    nTriangles = info->nTriangles;
    memcpy(triangles, info->triangles, 15 * sizeof(int));

    // Compute the interpolation parameter
    for (int i = 0; i < info->nVertices; i++) {
		float f0 = (float)values[info->edgeStart[i]];
		float f1 = (float)values[info->edgeEnd[i]];
		float diff = f1 - f0;
        param[i] =  diff==0 ? .5 : (level - f0) / diff;
    }
}


bool VolGridMesh::iterate(OcTreeTraversalNode *ocTravNode) {

	int i, j;
	float xyz[8][3];

	VolGridTraversalNode *travNode = (VolGridTraversalNode *) ocTravNode;
    VolGridNode *node = (VolGridNode *) travNode->node;

	// Determine if child is a leaf node at maximum depth.
	if (travNode->depth == this->grid->getMaxDepth() && ((VolGridNode *) (travNode->node))->nodeType == LEAF &&
		((VolGridNode *) (travNode->node))->valid == true) {
		
		// Determine coordinates of the vertices.
		for (i = 0; i < 8; i++) {			
			// set the correct x, y and z coordinates
			xyz[i][0] = travNode->center[0] + travNode->length * octreeChildCenterOffsetInt[i][0];
			xyz[i][1] = travNode->center[1] + travNode->length * octreeChildCenterOffsetInt[i][1];
			xyz[i][2] = travNode->center[2] + travNode->length * octreeChildCenterOffsetInt[i][2];
		}

		// Call marching cubes algo for current cell
		int nVertices;
		int edgeStart[12], edgeEnd[12];
		float param[12];
		int nTriangles;
		int triangles[5][3];
		float f[8];
		
		for (j = 0; j < 8; j++) {
			f[j] = *(node->distanceValues[j]);
		}

		this->marchingCubes(f, 0.0f, nVertices, edgeStart, edgeEnd, param, nTriangles, triangles);

		// Interpolate surface vertices
		float vertex[12][3];
		for (int v = 0; v < nVertices; v++) {
			int index0 = edgeStart[v];
			int index1 = edgeEnd[v];
			float t = param[v];
			vertex[v][0] = (1 - t) * xyz[index0][0] + t * xyz[index1][0];
			vertex[v][1] = (1 - t) * xyz[index0][1] + t * xyz[index1][1];
			vertex[v][2] = (1 - t) * xyz[index0][2] + t * xyz[index1][2];
		}

		// Output triangles in RAW triangle file format
		// ax, ay, az, bx, by, bz, cx, cy, cz
		for (int t = 0; t < nTriangles; t++) {
			int a = triangles[t][0];
			int b = triangles[t][1];
			int c = triangles[t][2];
			
			*this->ts << vertex[c][0] << " " << vertex[c][1] << " " << vertex[c][2] << " " 
					  << vertex[b][0] << " " << vertex[b][1] << " " << vertex[b][2] << " "
					  << vertex[a][0] << " " << vertex[a][1] << " " << vertex[a][2] << "\r\n";
		}
	}

	return true;
}


void VolGridMesh::build() {

	// Open file dialog box
    QFile fp(this->fileName);
    if (!fp.open( IO_WriteOnly)) {
        QMessageBox::warning(this->widget,"I/O Error", QString("The file could not be opened for writing.\n\n") + fileName);
		return;
    }

    this->ts = new QTextStream(&fp);
	this->grid->traverse(this);
	delete ts;
	fp.flush();
	fp.close();
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
